import importlib
import os
import sys
import datetime
import traceback
import re
import glob
import shutil
import argparse
import logging
import importlib

import xframe
#from xframe.library.pythonLibrary import xprint
def xprint(m):
    print(m)
    

def import_settings():
    from xframe import settings
    #xframe.__setattr__('settings',settings)
    
def setup_logging():
    from xframe.settings import general
    from xframe import logger
    log = logger.setup_custom_logger('root',general.loglevel)
    globals().update({'log':log})

# xframe modules to load at import 
modules_to_import  = ['Multiprocessing','database','control.Control',['library','lib'],'presenters']
def startup_imports():
    for module in modules_to_import:
        try:
            if isinstance(module,(list,tuple)):
                module_name = 'xframe.'+module[0]
                xframe.__setattr__(module[1], importlib.import_module(module_name))
            else:
                splitted_name = module.split('.')
                module_name = 'xframe.'+module
                if len(splitted_name) > 1:                
                    xframe.__setattr__(splitted_name[-1], importlib.import_module(module_name))
                else:
                    xframe.__setattr__(module,importlib.import_module(module_name))
            #log.info('loaded {}'.format(module))
                    
        except Exception as e:
            traceback.print_exc()
            log.error('Caught exeption while loading {} with content: {}'.format(module_name,e))



def create_dependency_dict():
    external_dependency_dict = {
        ##### math stuff #####
        'gsl':[(('gsl_plugin','GSLPlugin','pygsl'),('xframe.lib.math','gsl'),())],
        'flt':[(('flt_plugin','LegendreTransform','flt'),('xframe.lib.math','leg_trf'),())],
        'shtns':[(('shtns_plugin','sh','shtns'),('xframe.lib.math','shtns'),())],
        'persistent_homology':[(('persistent_homology','PersistentHomologyPlugin','None'),('xframe.lib.math','PeakDetector'),())],
        'pysofft':[(('soft_plugin','Soft','pysofft'),('xframe.lib.math','Soft'),())],
        ##### multiprocessing & GPU #####
        'shared_array':[('SharedArray',('xframe.lib.python','sa'),()),
                        ('SharedArray',('xframe.Multiprocessing','sa'),())],
        'opencl':[(('openCL_plugin','OpenClPlugin','pyopencl'),('xframe.Multiprocessing','openCL_plugin'),())],
        'psutil':[(('psutil_plugin','PsutilPlugin','psutil'),('xframe.Multiprocessing','psutil_plugin'),())],
        ###### database plugins ######
        'vtk':[(('vtk_plugin','vtkSaver','vtk'),('xframe.database.database','VTK_saver'),())],
        'pdb':[(('pdb_plugin','ProteinDB','pdb_eda'),('xframe.database.database','PDB_loader'),())],
        'hdf5':[(('hdf5_plugin','HDF5_DB','h5py'),('xframe.database.database','HDF5_access','as_object'),())],
        'yaml':[(('yaml_plugin','YAML_access','ruamel_yaml'),('xframe.database.database','YAML_access'),())],
        'opencv':[(('cv_plugin','CV_Plugin','opencv-python'),('xframe.database.database','OpenCV_access'),()),
                  (('cv_plugin','CV_Plugin','opencv-python'),('xframe.presenters.openCVPresenter','CV_Plugin'),())],
        ###### presenter plugins ######        
        'matplotlib':[(('matplotlib_plugin','matplotlib','matplotlib'),('xframe.presenters.matplotlibPresenter','matplotlib'),('xframe.presenters.matplotlibPresenter.depencency_injection_hook_mpl',))],
        'matplotlib_toolkit':[(('mpl_toolkits_plugin','mpl_toolkits','matplotlib'),('xframe.presenters.matplotlibPresenter','matplotlib'),('xframe.presenters.matplotlibPresenter.dependency_injection_hook_mpl_toolkits',))],
        ###### cmd ######
        'click':[(('click_plugin','ClickPlugin','click'),(False,False,'as_object'),())],
    }
    return external_dependency_dict
    
def _inject_dependency(dependency_name,reload=False):
    log=logging.getLogger('root')
    log.info(f'Dependency injecting {dependency_name}.')
    dependency_list = xframe._external_dependency_lookup.get(dependency_name,False)
    if isinstance(dependency_list,bool):
        log.warning(f'xFrame does not know the dependency {dependency_name}, skipping. Known external dependencies are {xframe._external_dependency_lookup.keys()}')
        return
    ## Failing to import does not generate an error message or raise an error.
    ## If later a routine is called that would require a
    ## depencency it is called against an abstract calss that throws a corresponding not imported error.
    for import_source ,import_destination,injection_hooks in dependency_list:
        if isinstance(import_source,(list,tuple)):
            import_name='xframe.externalLibraries.'+ import_source[0]
            tmp = importlib.import_module(import_name)
            dependency = tmp.__getattribute__(import_source[1])                
        else:
            #import_name='xframe.externalLibraries.'+ import_source
            dependency = importlib.import_module(import_source)
        if reload:
            dependency = importlib.reload(dependency)
        if len(import_destination)==3:
            if import_destination[2]=='as_object':
                dependency = dependency()
        if not isinstance(import_destination[0],bool):
            setattr(eval(import_destination[0]),import_destination[1],dependency)
            if len(injection_hooks)>0:
                for hook in injection_hooks:
                    eval(hook)()
    return dependency
            
# External packages to be dependency injected.
# has to be callde after startup_imports
def dependency_injection():
    if not xframe.settings.general.lazy_denpendency_injection:
        for key in xframe._external_dependency_name_.keys():
            if key != 'pysofft':
                _inject_dependency(key)

def dependency_inject_SOFT():
    # separate since it uses numba and requires a bit of time to import.
    try:
        from xframe.externalLibraries.soft_plugin import Soft
        xframe.lib.math.Soft=Soft
    except Exception as e:
        log.error(f'loading PySOFT dependency failed with error {e}')

        
def setup_default_database():
    xframe.database.default = xframe.database.database.DefaultDB(**xframe.settings.general.dict()['IO'])


def lookup_projexp(path_list):
    _dict = {}
    #print(f'{path_list}')
    for path in path_list:
        try:
            names = next(os.walk(path))[1]
            for name in names:
                if (name[0]!='_') and (name not in _dict):
                    _dict[name] = os.path.join(path,name+'/')
        except StopIteration as e:
            pass
    return _dict
## look for projects
def lookup_projects():
    opt = xframe.settings.general
    db = xframe.database.default

    project_paths = [db.get_path(fname,is_file=False) for fname in opt.project_folders]
    project_dict = lookup_projexp(project_paths)
    if len(project_dict) == 0:
        log.warning('No xFrame projects found at {}. Check the general settings project path option.'.format(project_paths))
    return project_dict    
def setup_projects():
    xframe.known_projects = lookup_projects()
    xframe.__setattr__('projects',importlib.import_module('xframe.projects'))

## look for experiments
def lookup_experiments():
    opt = xframe.settings.general
    db = xframe.database.default
    #log.info('looking for experiments at {}'.format(opt.experiment_paths))    
    experiment_paths = [db.get_path(fname,is_file=False) for fname in opt.experiment_folders]
    experiment_dict = lookup_projexp(experiment_paths)
    if len(experiment_dict) == 0:
        log.warning('No xFrame experiment found at {}. Check the general settings experiment path option.'.format(experiment_paths))
    return experiment_dict    

def setup_experiments():
    xframe.known_experiments = lookup_experiments()
    xframe.__setattr__('experiments',importlib.import_module('xframe.experiments'))


def lookup_workers(project_path):
    from xframe.settings import general as opt
    file_paths = glob.glob(os.path.join(project_path,'*.py'))
    file_names = tuple(os.path.basename(p) for p in file_paths)
    #print(f'worker candidates = {file_names}')
    #print(f'worker regexpr = {opt.worker_regexpr}')
    pattern = re.compile(opt.worker_regexpr)
    worker_names = [os.path.splitext(name)[0] for name in file_names if pattern.match(name)]
    #print(f'workers = {worker_names}')
    return worker_names

def _parse_project_name(name,worker_name=False):
    known_projects = xframe.known_projects
    project_paths = [xframe.database.default.get_path(name,is_file=False) for name in xframe.settings.general.project_folders]
    if not name in known_projects:
        raise AssertionError('Project_Name {} was not found at {}. Known projects are {}'.format(name,[path + name for path in project_paths],known_projects.keys()))
    project_path = known_projects[name]
    if isinstance(worker_name,str):        
        worker_path = os.path.join(project_path,worker_name+'.py')
        worker_exists = os.path.exists(worker_path)
    else:
        worker_exists = False
    if not worker_exists:
        raise AssertionError('Worker {} for Project {} was not found at {}'.format(worker_name,name,worker_path))            
    return project_path,name,worker_name

def _parse_experiment_name(name):
    known_experiments = xframe.known_experiments
    if not name in known_experiments:
        raise AssertionError('Experiment "{}" was not found at {}. Known experiment modules are "{}"'.format(name,[path + name for path in xframe.settings.general.experiment_paths],known_experiments,keys()))
    experiment_path = known_experiments[name]
    return experiment_path,name

def _load_db(name,db_name,opt,target='project'):
    if target=='experiment':
        db_module_pypath = 'xframe.experiments.'+name+'._database_'  
    else:
        db_module_pypath = 'xframe.projects.'+name+'._database_'
        
    try:      
        db = getattr(importlib.import_module(db_module_pypath),db_name)(**opt)
    except (AttributeError, ModuleNotFoundError) as e:
        log.info('Could not instanciate database class {} in project {} use default database instead. Recived the following error during initialization: \n {}'.format(db_name,name,e))
        log.debug(traceback.format_exc())
        db = xframe.database.database.DefaultDB(**opt)
    return db
   

def select_project(name=False,worker_name=False,project_settings=False,ignore_file_not_found=False):
    #sys.path.append(opt.project_paths)
    settings = xframe.settings
    database = xframe.database
    
    proj_settings_name = project_settings
    if isinstance(name,bool):
        project_settings = {}
        project_settings_raw = {}
        project_path = ''
        name = ''
        worker_name = ''
    else:        
        project_path,name,worker_name = _parse_project_name(name,worker_name)
        project_settings,project_settings_raw = database.default.load('settings',project_path = project_path,worker_name = worker_name,settings_file_name=project_settings,ignore_file_not_found=ignore_file_not_found)

        project_settings_raw = project_settings_raw.dict()
    settings._update_settings_on_project_select(project_settings,project_settings_raw)
    #settings.project = project_settings
    #settings.raw_project= project_settings_raw
    if 'IO' in project_settings:
        project_IO_settings=project_settings.dict().pop('IO')
    else:
        project_IO_settings = {}
    project_db = _load_db(name,settings.general.default_project_db_name,project_IO_settings)
    database.project = project_db
    xframe._project_worker_module_name = _get_worker_module_name(name,worker_name)
    
def select_experiment(exp_name=False,exp_settings=False,ignore_file_not_found=False):
    settings = xframe.settings
    database = xframe.database
    controller = xframe.controller
    
    exp_settings_name = exp_settings
    if isinstance(exp_name,str):
        exp_path,exp_name = _parse_experiment_name(exp_name)
        exp_settings,raw_exp_settings=database.default.load('settings',project_path = exp_path,settings_file_name=exp_settings,target='experiment',ignore_file_not_found=ignore_file_not_found)
        raw_exp_settings = raw_exp_settings.dict()
        settings._update_settings_on_experiment_select(exp_settings,raw_exp_settings)
        #settings.experiment = exp_settings
        #settings.raw_experiment = raw_exp_settings
        if 'IO' in exp_settings:
            exp_IO_settings = exp_settings.dict().pop('IO')
        else:
            exp_IO_settings = {}
        exp_db = _load_db(exp_name,settings.general.default_experiment_db_name,exp_IO_settings,target='experiment')
        database.experiment = exp_db
        xframe._experiment_module_name = _get_worker_module_name(exp_name,is_experiment=True)

def import_selected_project(update_worker=True):
    settings = xframe.settings
    database = xframe.database
    controller = xframe.controller    
    
    if update_worker:
        settings.raw_project = database.default.format_settings(settings.project)
        
    module_name=xframe._project_worker_module_name
    worker_module_imported = module_name in sys.modules
    worker_instance_exists = isinstance(xframe.project_worker,xframe.interfaces.ProjectWorkerInterface)
    needs_worker_creation = (not worker_module_imported) or (not worker_instance_exists) or update_worker
    #xprint(f'importing {module_name} needs worker update = {needs_worker_creation}')
    if needs_worker_creation:        
        try:
            if module_name in sys.modules:
                #xprint('reloading')
                worker_module = importlib.reload(sys.modules[module_name])
            else:
                #xprint('importing')
                worker_module = importlib.import_module(module_name)
            #xprint('instanciating')
            worker_instance = getattr(worker_module,settings.general.default_project_worker_name)()
            xframe.project_worker =  worker_instance
            controller.project_worker = worker_instance
        except Exception as e:
            log.error(f'Could not import project worker {module_name},\n with error {e}\n {traceback.format_exc()}\n Terminating.')
            sys.exit()
            
def import_selected_experiment(update_worker=True):
    if isinstance(xframe._experiment_module_name,str):
        settings = xframe.settings
        database = xframe.database
        controller = xframe.controller    
        
        if update_worker:
            settings.raw_experiment = database.default.format_settings(settings.experiment)
    
        module_name = xframe._experiment_module_name
        module_imported = module_name in sys.modules
        instance_exists = isinstance(xframe.experiment,xframe.interfaces.ExperimentWorkerInterface)
        needs_experiment_creation = (not module_imported) or (not instance_exists) or update_worker
        
        if needs_experiment_creation:        
            try:
                if module_name in sys.modules:
                    exp_module = importlib.reload(sys.modules[module_name])
                else:
                    exp_module = importlib.import_module(module_name)            
                exp_instance = getattr(exp_module,settings.general.default_experiment_worker_name)()
                xframe.experiment =  exp_instance
                controller.experiment_worker = exp_instance
            
            except Exception as e:
                log.error(f'Could not import experiment {module_name},\n with error {e}\n {traceback.format_exc()}\n Terminating.')
                sys.exit()

                
def _get_worker_module_name(name,worker=None,is_experiment=False):
    if is_experiment:
        worker_module_name = 'xframe.experiments.'+name+'.'+xframe.settings.general.default_experiment_module_name
    else:
        if isinstance(worker,str):
            worker_module_name = 'xframe.projects.'+name+'.'+worker
        else:
            raise AssertionError(f"Project worker not specified.")
    return worker_module_name
    


def select_and_run(project='project', project_worker='project_worker',project_settings='settings_name',exp_name=False,exp_settings=False, update_worker=False,oneshot=False):
    select_project(name = project, worker_name = project_worker, project_settings = project_settings)
    select_experiment(exp_name = exp_name, exp_settings = exp_settings)
    import_selected_project(update_worker=update_worker)
    import_selected_experiment(update_worker=update_worker)
    xframe.lib.python.measureTime(xframe.controller.run)(oneshot=oneshot)

def run(update_worker = False,oneshot=False):
    import_selected_project(update_worker=update_worker)
    import_selected_experiment(update_worker=update_worker)
    xframe.controller.run(oneshot=oneshot)


class DefaultProjectArgparser:
    project_description=''
    def __init__(self):
        self.settings_parser = argparse.ArgumentParser(add_help=False)
        self.settings_parser.add_argument('settings',metavar='SETTINGS_NAME',default=False,type=str,help='Name of settings file to be used')
    def add_parser(self,subparser,project_name,project_path):
        worker_names = lookup_workers(project_path)
        #log.info()
        description = self.project_description
        project_parser = subparser.add_parser(project_name, description=description)
        project_subparser = project_parser.add_subparsers(help='Available Workers',dest = 'worker')
        for worker in worker_names:
            add_worker_parser = getattr(self,'add_'+worker+'_argparser',self.add_default_worker_parser)
            #print(add_worker_parser)
            add_worker_parser(project_subparser,worker)
    
    def add_default_worker_parser(self,subparser,worker):
        subparser.add_parser(worker, description='No description available', help=f'No help available',parents=[self.settings_parser])

    def process_arguments(self,args):
        pass

class DefaultProjectClick:
    project_description = ''
    short_description = ''
    worker_help = {}
    def add_click(self,group,click,project_name,project_path):
        parsed_results= {}
        self.worker_names = lookup_workers(project_path)
        self.project_name = project_name
        description = self.project_description
        add_project_click = getattr(self,'add_'+project_name,self.add_default_project_click)
        worker_group = add_project_click(group,click)
        for worker in self.worker_names:
            try:
                add_worker_parser = getattr(self,'add_worker_'+worker,self.add_default_worker_click)
                add_worker_parser(worker_group,click,worker)
            except Exception as e:
                log.error(e)
    
    def add_default_worker_click(self,group,click,worker_name,help='',short_help='',epilog=''):
        @group.command(worker_name,help=help,short_help=short_help,epilog=epilog)
        @click.argument('settings_name',required=False)
        def worker(settings_name=False,**kwargs):
            log.info(f'start: {worker_name}')
            #xprint(f'selecting worker {worker_name}')
            if settings_name is None:
                #print('no settings provided')
                select_project(name=self.project_name,worker_name=worker_name)
            else:
                #print('settings provided')
                select_project(name=self.project_name,worker_name=worker_name,project_settings=settings_name)
            xframe.library.pythonLibrary.measureTime(run)(oneshot=False)
    def add_default_project_click(self,group,click):
        @group.group(self.project_name,chain=True,help = self.project_description, short_help= self.short_description)
        def project():
            pass
        return project
    def process_arguments(self,args):
        pass

    
def setup_home(home,silent=False):
    db = xframe.database.default
    if os.path.expanduser(home) != db.get_path('home',is_file=False):
        if home[-1]!='/':
            home+='/'
        new_home_exists = os.path.exists(os.path.expanduser(home))
        if not new_home_exists:
            response = 'a'
            if not silent:
                while response not in 'yYnN':                
                    response = input(f'Folder "{home}" does not exist. Do you want to create the new folder [y/n]? ')
                if response in 'nN':
                    xframe.lib.python.xprint('abborting!')
                    sys.exit()
        log.info(f'Changing home from {db.get_path("home",is_file=False)} to {home}')
        settings_entry = f'home_folder = "{os.path.expanduser(home)}"\n'
        
        #Creating config file in .xframe pointing to the new home directory
        if os.path.exists(db.get_path('config')):
            lines = db.load('config',as_text= True)
            home_folder_found  =False
            for _id,line in enumerate(lines):
                if line[:11]=='home_folder':
                    lines[_id]= settings_entry
                    home_folder_found = True
                    break
            if not home_folder_found:
                lines = [settings_entry]+lines
            config= lines
        else:
            config = [settings_entry]        
        db.save('config',config)
        xframe.settings.general.home_folder = home
        xframe.settings.general.IO.folders.home=home
        db.update_folders_and_files(**xframe.settings.general.IO.dict())        

    log.info('symlinking projects and experiments from install folder to home.')
    db.create_folders('home')
    distributed_projects = lookup_projexp([db.get_path('install_projects',is_file=False)])
    distributed_experiments = lookup_projexp([db.get_path('install_experiments',is_file=False)])
    log.info(f'distributed_projects = {distributed_projects}')
    for project,install_path in distributed_projects.items():
        dest = os.path.join(db.get_path('projects',is_file=False),project)
        log.info(f'source = {install_path} target = {dest}')
        db.create_symlink(install_path,dest,is_file=False)        
    for experiment,install_path in distributed_experiments.items():
        dest = os.path.join(db.get_path('experiments',is_file=False),experiment)
        log.info(f'source = {install_path} target = {dest}')
        db.create_symlink(install_path,dest,is_file=False)
    #db.create_symlink('experiments_install','experiments',is_file=False)
    #db.create_symlink('projects_install','projects',is_file=False)
    log.info('symlinking default settings.')
    for project,project_path in distributed_projects.items():
        for worker in lookup_workers(project_path):            
            project_path_modifier = {'project':project,'worker':worker}
            db.create_folders('settings_project',path_modifiers=project_path_modifier)
            settings_install_path = db.get_path('settings_install_project',path_modifiers=project_path_modifier,is_file=False)
            home_settings_path = db.get_path('settings_project',path_modifiers=project_path_modifier,is_file=False)
            default_settings_files = glob.glob(os.path.join(settings_install_path,'default_*.yaml'))
            for default in default_settings_files:
                target_file = os.path.join(home_settings_path,os.path.basename(default))
                db.create_symlink(default,target_file)
                log.info(f'source = {default} target = {target_file}')
    log.info('copy tutorial settings')
    for project,project_path in distributed_projects.items():
        for worker in lookup_workers(project_path):            
            project_path_modifier = {'project':project,'worker':worker}
            db.create_folders('settings_project',path_modifiers=project_path_modifier)
            settings_install_path = db.get_path('settings_install_project',path_modifiers=project_path_modifier,is_file=False)
            home_settings_path = db.get_path('settings_project',path_modifiers=project_path_modifier,is_file=False)
            tutorial_settings_files = glob.glob(os.path.join(settings_install_path,'tutorial*.yaml'))
            for tutorial in tutorial_settings_files:
                target_file = os.path.join(home_settings_path,os.path.basename(tutorial))
                db.copy(tutorial,target_file)
                log.info(f'source = {tutorial} target = {target_file}')
                #print(f'source = {default} target = {target_file}')
    log.info('symlink default experiment settings')
    for exp,exp_path in xframe.known_experiments.items():
        #xframe.database.default.create_folders('user_experiment_settings',path_modifiers={'experiment':exp})
        exp_modifier = {'experiment':exp}
        db.create_folders('settings_experiment',path_modifiers=exp_modifier)
        settings_install_path = db.get_path('settings_install_experiment',path_modifiers=exp_modifier,is_file=False)
        home_settings_path = db.get_path('settings_experiment',path_modifiers=exp_modifier,is_file=False)
        
        default_settings_files = glob.glob(os.path.join(settings_install_path,'default_*.yaml'))
        for default in default_settings_files:
            target_file = os.path.join(home_settings_path,os.path.basename(default))
            log.info(f'source = {default} target = {target_file}')
            db.create_symlink(default,target_file)
            
    log.info('copy experiment tutorial settings')
    for exp,exp_path in xframe.known_experiments.items():
        #xframe.database.default.create_folders('user_experiment_settings',path_modifiers={'experiment':exp})
        exp_modifier = {'experiment':exp}
        db.create_folders('settings_experiment',path_modifiers=exp_modifier)
        settings_install_path = db.get_path('settings_install_experiment',path_modifiers=exp_modifier,is_file=False)
        home_settings_path = db.get_path('settings_experiment',path_modifiers=exp_modifier,is_file=False)        
        tutorial_settings_files = glob.glob(os.path.join(settings_install_path,'tutorial*.yaml'))
        for tutorial in tutorial_settings_files:
            target_file = os.path.join(home_settings_path,os.path.basename(tutorial))
            db.copy(tutorial,target_file)
            log.info(f'source = {tutorial} target = {target_file}')
            #print(f'source = {default} target = {target_file}')
    xprint(f'Created xframe home at: {os.path.expanduser(home)}')

 
def change_home(new_home_path=None):
    if new_home_path is None:
        new_home_path = xframe.settings.general.default_folder
    if new_home_path[-1]!='/':
        new_home_path+='/'
        
    if not os.path.exists(new_home_path):
        setup_home(new_home_path,silent=True)
    else:
        log.info(f'Changing home from {db.get_path("home",is_file=False)} to {home}')
        settings_entry = f'home_folder = "{os.path.expanduser(home)}"\n'        
        #Creating config file in .xframe pointing to the new home directory
        if os.path.exists(db.get_path('config')):
            lines = db.load('config',as_text= True)
            home_folder_found  = False
            for _id,line in enumerate(lines):
                if line[:11]=='home_folder':
                    lines[_id]= settings_entry
                    home_folder_found = True
                    break
            if not home_folder_found:
                lines = [settings_entry]+lines
            config= lines
        else:
            config = [settings_entry]        
        db.save('config',config)
        xframe.settings.general.home_folder = home
        xframe.settings.general.IO.folders.home=home
        db.update_folders_and_files(**xframe.settings.general.IO.dict())                
    reload_home()

    
def reload_home():
    global xframe
    xframe = importlib.reload(xframe)
    xframe.projects = importlib.reload(xframe.projects)
    xframe.experiments = importlib.reload(xframe.experiments)
        
        



import time
import numpy as np
from scipy.interpolate import griddata
import logging
import traceback
import warnings

log=logging.getLogger('root')

from xframe.library.physicsLibrary import ewald_sphere_theta
from xframe.library.gridLibrary import GridFactory
from xframe.library.mathLibrary import cartesian_to_spherical
from xframe.library.mathLibrary import spherical_to_cartesian
from xframe.library import units


#very bad saxs is average not integral...
def calc_saxs_bad(mean,phis,thetas):
    log.info('nonzero entries in mean for saxs = {}'.format(np.sum(mean!=0)))
    saxs= np.trapz(mean*np.sin(thetas)[:,None],x = phis, axis = 1)
    #saxs= np.trapz(mean,x = phis, axis = 1)
    return saxs

def calc_saxs(mean,mask):
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore')        
        saxs= np.mean(mean, axis = 1, where = mask)
    nan_mask = np.isnan(saxs)
    if nan_mask.any():
        saxs[nan_mask] = 0
    #saxs= np.trapz(mean,x = phis, axis = 1)
    return saxs

def calc_saxs_slow(mean,mask):
    saxs = np.zeros(len(mean))
    for q_id in range(len(mean)):
        q_mask = mask[q_id]
        _sum = np.sum(mean[q_id],where = q_mask)
        n_points = np.sum(q_mask)
        if n_points != 0:
            saxs[q_id] = _sum/n_points
        else:
            saxs[q_id] = 0

    #saxs= np.trapz(mean,x = phis, axis = 1)
    return saxs




import os
import logging
import traceback
import numpy as np

from itertools import repeat

from xframe.control import communicators as commModules
from xframe.interfaces import ProjectWorkerInterface,ExperimentWorkerInterface
from xframe.library import pythonLibrary as pyLib
from xframe.library.pythonLibrary import DictNamespace, xprint
from xframe.Multiprocessing import get_local_cpu_count as cpu_count
from xframe import settings
from xframe import database
from xframe.Multiprocessing import process_mp_request, MPMode
from xframe import Multiprocessing
#get_global_number_of_processes = Multiprocessing.get_global_number_of_processes

#import warnings
#warnings.filters.append(('ignore',None,UserWarning,None,0))

log=logging.getLogger('root')




class Controller:
    def __init__(self,project_worker=False,experiment_worker=False):       
        self._project_worker = project_worker
        self._experiment_worker = experiment_worker
        self.control_worker=ControlWorker(project_worker = project_worker,experiment_worker=experiment_worker)        
        Multiprocessing.comm_module = commModules.SingleProcessCommunictionAnalysis(control_worker = self.control_worker)
        self.job='not jet chosen'
        
    @property
    def project_worker(self):
        return self._project_worker
    @property
    def experiment_worker(self):
        return self._experiment_worker
    @project_worker.setter
    def project_worker(self,worker):
        try:
            #log.info('trying to set analysis worker in COntroller')
            #assert isinstance(worker,AnalysisWorker),'analysis_worker has to inherit from {} but the given object {} does not.'.format(AnalysisWorker,worker)
            self._project_worker = worker
            self.control_worker.project_worker = worker
            #log.info(self._analysis_worker)
        except AssertionError as e:
            log.error(e)            
    @experiment_worker.setter
    def experiment_worker(self, worker):
        try:
            #assert isinstance(worker,Experiment_WorkerWorkerInterface),'experiment_worker_worker has to inherit from {} but the given object {} does not.'.format(Experiment_WorkerWorkerInterface,worker)
            self._experiment_worker = worker
            self.control_worker.experiment_worker = worker
        except AssertionError as e:
            log.error(e)
            
            
            
    def run(self,oneshot=False,restart_control_worker=False):
        settings._save_settings_on_controller_run()
        if restart_control_worker:
            # restart allows the control_workers to reload settings and change from the default
            # ones to the settings provided by the analysis worker. (e.g for profiling)
            self.control_worker.restart_working()
        try:
            if not self.control_worker.are_workers_running:
                self.control_worker.start_working()
                
            job=self.choose_job()
            self.job = job
            xprint('\n ------- Start {} ------- \n'.format(job))
            result=job.run()
            xprint('\n ------- Finished {} ------- \n'.format(job))
        except Exception as e:
            traceback.print_exc()
            result = None
            log.error('The follwoing exception occured during, or directly before, job execution:\n {}'.format(e))
            
        if oneshot:
            del(self.control_worker.gpu_manager)
            self.control_worker.stop_working()
            
        if result != None:
            return result

    def choose_job(self):
        n_running_nodes = 1 #get_global_number_of_processes()
        if n_running_nodes==1:
            job=self.choose_job_single_node()
        else:
            job=self.choose_job_multi_node()
        return job
            
    def choose_job_single_node(self):
        return self.project_worker
    
    def choose_job_multi_node(self):
        return self.project_worker

    
class ControlWorker:
    def __init__(self,project_worker=False,experiment_worker=False,start_working=True):
        self.project_worker=project_worker
        self.experiment_worker=experiment_worker
        #self.dataSetsToAnalizeByID=dataSetsToAnalizeByID
        #self.remainingIDs=dataSetsToAnalizeByID.copy()
        #manager=Manager()       
        self.gpu_manager = Multiprocessing.GpuProcessManager()
        self.workers = []
        if start_working:
            self.start_working()

    def are_workers_running(self):
        some_workers_are_running = False
        for worker in self.workers:
            if worker.process.is_alive():
                some_workers_are_running = True
                break
        return some_workers_are_running
    def start_working(self):        
        if not self.are_workers_running():
            #log.info('run {} controll workers'.format(len(self.worker_processes)))
            if settings.general.n_control_workers!=0:
                _,workers = self.process_mp_request_single_node(self.work,is_daemon = True,return_process_objects = True,n_processes = settings.general.n_control_workers,create_queues = True)                
                self.workers = workers
        else:
            log.error('Some workers are still running can not start them again. Call stop_working or restart_working to chancle the old ones.')
    def stop_working(self):
        for worker in self.workers:
            worker.queue.put(("terminate",0,True))
            worker.process.join()
        self.workers = []
        
    def restart_working(self):
        if self.are_workers_running():
            self.stop_working()
        self.start_working()
        log.info('Run with {} control workers'.format(len(self.workers)))
        

    def work(self,**kwargs):
        q = kwargs['queue']
        #if not isinstance(self.gpu_manager,Multiprocessing.GpuProcessManager):
        #    print('Something went wrong, starting new GpuProcessManager!')
        #    self.gpu_manager = Multiprocessing.GpuProcessManager()
        run_profiling=settings.project.get('profiling',DictNamespace(enable=False)).enable
        if run_profiling:
            process_id = settings.project.profiling.gpu_worker_id
            #log.info('controler name = {}'.format(Multiprocessing.get_process_name()))
            if Multiprocessing.get_process_name() == -1*np.abs(process_id):
                try:
                    db = database.project
                    path = db.get_reconstruction_path()
                except Exception:                    
                    db = database.default
                    path = db.get_path('base',is_file=False)
                db.create_path_if_nonexistent(path)
                path += "control_worker_{}.stats".format(process_id)
                log.info('profile path = {}'.format(path))
                import cProfile
                profiler = cProfile.Profile()
                profiler.enable()
                self.request_loop(q)
                profiler.disable()
                profiler.dump_stats(path)
            else:
                self.request_loop(q)
        else:
            self.request_loop(q)
            
    def request_loop(self,q):
        try:
            Multiprocessing.create_openCL_context()
            #log.info('context created in process {}'.format(Multiprocessing.get_process_id))
            gpu_process_request=self.gpu_manager.process_request
            #log.info('worker pid = {}'.format(os.getpid()))
            while True:
                request_type,cpu_id,data = q.get()            
                if request_type == "terminate":
                    #log.info('Request {} from process {} to {}'.format(request_type,cpu_id,Multiprocessing.get_process_name()))
                    break
                elif request_type == 'ping':
                    worker_id = Multiprocessing.get_process_name()
                    message = ('ping',worker_id)
                    Multiprocessing.put_to_process_queue(cpu_id,message)
                else:
                    gpu_process_request(request_type,cpu_id,data)
        finally:
            self.exit_request_loop()
            
        
    def exit_request_loop(self):
        #print('deleting gpu_manager in P {}'.format(Multiprocessing.get_process_name()))
        self.gpu_manager.release_shared_memory()
        self.gpu_manager = Multiprocessing.GpuProcessManager()
        
    def select_worker_queue(self):
        p_id = Multiprocessing.get_process_id()
        return self.queues[p_id%len(self.queues)]

    def select_control_worker(self):
        p_id = Multiprocessing.get_process_name()
        worker = self.workers[p_id%len(self.workers)]
        #log.info('reconstruction process {} selected control worker {}'.format(p_id,worker.name))
        return worker
    def processDataRequest_SingleProcess(self,opt):
        data_generator = self.experiment_worker.get_data(opt)
        return data_generator

    def process_geometry_request_single_thread(self,**kwargs):
        geometry=self.experiment_worker.get_geometry(**kwargs)
        return geometry

    def process_exp_info_request_single_thread(self):
        return self.experiment_worker.info

    def process_mp_request_single_node(self,*args,**kwargs):
        return process_mp_request(*args,**kwargs)
        



import pygsl.testing.sf as sf
import numpy as np
from xframe.library.interfaces import GSLInterface

class GSLPlugin(GSLInterface):
    lib = sf
    @staticmethod
    def legendre_sphPlm_array(l_max,m_max,xs,return_orders = False,sorted_by_l = False):
        '''
        Returns array of all normalized associated Legendre coefficients $\overline{P}^m_l$ with orders $l\leql_{max}$.
        :param l_max: Maximum Order to consider
        :type l_max: int
        :param xs: Array or value of points at which to evaluate the associated legendre polynomials
        :type xs: numpy.ndarray or float
        :param return_orders: Wether or not to return the order arrays l and m for which values where calculated. 
        :type return_orders: bool
        :param sort_by_l: If true output is ordered like ($P^0_0,P^0_1,P^1_1,P^0_2,P^1_2,P^2_2,\ldots$) and otherwise the output is ordered by m (i.e.$ P^0_0,\ldots,P^0_{l_max},P^1_1,\ldots,P^0_{l_max},\ldots$)
        :type sort_by_l: bool
        :return: Array with coefficients. If return_orders is true, then the order arrays are returned as well.
        :rtype: numpy.ndarray or tuple
        '''
        if sorted_by_l:
            return GSLPlugin.legendre_sphPlm_array_l(l_max,m_max,xs,return_orders = return_orders)
        else:
            return GSLPlugin.legendre_sphPlm_array_m(l_max,m_max,xs,return_orders = return_orders)
    @staticmethod
    def legendre_sphPlm_array_l(l_max,m_max,xs,return_orders = False):
        ls = np.arange(l_max+1)
        ms = np.concatenate(tuple(np.arange(l+1)[:m_max+1] for l in ls))
        repeated_ls = np.concatenate(tuple(np.full(min(l+1,m_max+1),l) for l in ls))
        xs = np.atleast_1d(xs)
        values = sf.legendre_sphPlm(repeated_ls[:,None],ms[:,None],xs[None,:])
        if return_orders:
            return (np.squeeze(values),repeated_ls,ms)
        else:
            return np.squeeze(values)
    @staticmethod
    def legendre_sphPlm_array_m(l_max,m_max,xs,return_orders = False):        
        ms = np.arange(m_max+1)
        ls = np.concatenate(tuple(np.arange(m,l_max+1) for m in ms))
        repeated_ms = np.concatenate(tuple(np.full(max(0,l_max+1-m),m) for m in ms))
        xs = np.atleast_1d(xs)
        values = sf.legendre_sphPlm(ls[:,None],repeated_ms[:,None],xs[None,:])
        if return_orders:
            return (np.squeeze(values),ls,repeated_ms)
        else:
            return np.squeeze(values)

    @staticmethod
    def legendre_sphPlm_array_single_m(l_max,m,xs,return_orders = False):
        ls = np.arange(l_max+1)
        ms = np.full(len(ls),m)
        xs = np.atleast_1d(xs)
        values = sf.legendre_sphPlm(ls[:,None],ms[:,None],xs[None,:])
        if return_orders:
            return (np.squeeze(values),ls,ms)
        else:
            return np.squeeze(values)
    @staticmethod
    def legendre_sphPlm_array_single_l(l,l_max,xs,return_orders = False):
        ms = np.arange(0,l+1)
        ls = np.full(len(ms),l)
        xs = np.atleast_1d(xs)
        values = sf.legendre_sphPlm(l,ms[:,None],xs[None,:])
        
        if return_orders:
            return (np.squeeze(values),ls,ms)
        else:
            return np.squeeze(values)
        
    @staticmethod
    def bessel_jl(ls,xs):
        return sf.bessel_jl(ls,xs)

    @staticmethod
    def hyperg_2F1(a,b,c,z):        
        return sf.hyperg_2F1(a,b,c,z)


    @staticmethod
    def coupling_3j_e(two_ja, two_jb, two_jc, two_ma, two_mb, two_mc):        
        return sf.coupling_3j(two_ja, two_jb, two_jc, two_ma, two_mb, two_mc)

    @staticmethod
    def clebsch_gordan(J, j1, j2, m1, m2, M):        
        return (-1)**(j2-j1-M)*np.sqrt(2*J+1)*sf.coupling_3j(int(2*j1), int(2*j2), int(2*J), int(2*m1), int(2*m2), int(-2*M))

    @staticmethod
    def clebsch_gordan_2(J,j1, j2, m1, m2):
        M=m1+m2
        return (-1)**(np.abs(j2-j1-M))*np.sqrt(2*J+1)*sf.coupling_3j(2*j1, 2*j2, 2*J, 2*m1, 2*m2, -2*M)



import numpy as np
import sys
import importlib
import logging
log=logging.getLogger('root')
import pyopencl as cl
import abc
from xframe.Multiprocessing_interfaces import OpenClInterface
from xframe.Multiprocessing import get_process_name,master_name
from xframe.library.pythonLibrary import hash_numpy_in_tuple

int_type = np.int32

################################################

def _get_platforms(allow_master_process=False):
    '''
    By default prevents access to get_platforms for master process. This is because calls to cl.get_platforms() on the master call would cause all subsequent worker/children calls to cl.get_platforms to fail.
    So only set allow_master_process if you dont want to use multiprocessing.
    '''
    is_master_process = (get_process_name()==master_name)
    if (not is_master_process) or allow_master_process:
        platforms = cl.get_platforms()
    elif is_master_process:
        raise AssertionError("Creation of OpenCL context not allowed in master process.")
    return platforms

class ContextHandler:
    def __init__(self):
        self.ctx = None
        self.queue = None
        self.context_attached = False
    def attach_context(self,ctx,queue):
        try:
            assert (not self.context_attached), 'openCL context is already attached. Ignoring call to attach_context'
            self.ctx=ctx
            self.queue = queue
            self.context_attached = True
        except AssertionError as e:
            log.warning(e)
                    
class OpenClPlugin(OpenClInterface):
    cl_state = False
    contexts_created = False
    
    @classmethod
    def create_context(cls,allow_master_process=False):
        '''
        creates openCL context and queue for all gpus if they not already exist.        
        '''
        platform = _get_platforms(allow_master_process)
        my_gpu_devices = platform[0].get_devices(device_type=cl.device_type.GPU)
        cls.cl_state = []
        for device in my_gpu_devices:
            ctx = cl.Context(devices=[device])
            queue = cl.CommandQueue(ctx)        
            state = {}
            state['context']=ctx
            state['queue']=queue
            cls.cl_state.append(state)
        cls.contexts_created = True
        
    class ClFunction(ContextHandler):
        def __init__(self,func_data : dict, kernel,context_data=False):            
            super().__init__()
            if isinstance(context_data,(tuple,list)):
                self.attach_context(*context_data)
            self.kernel = kernel
            d = func_data            
            self.dict = d 
            self.name = d['name']
            self.dtypes = d['dtypes']
            self.shapes = d['shapes']
            self.arg_roles = d['arg_roles']
            self.const_inputs = d['const_inputs']
            self.global_range = self.read_ranges(d.get('global_range',None))
            self.local_range = self.read_ranges(d.get('local_range',None))
    
            self.n_args = len(self.arg_roles)
            #log.info("there are {} gpu arguments".format(self.n_args))
            self.input_dtypes = [self.dtypes[arg_id] for arg_id in range(self.n_args) if self.arg_roles[arg_id] == 'input']
            self.input_shapes = [self.shapes[arg_id] for arg_id in range(self.n_args) if self.arg_roles[arg_id] == 'input']
            self.input_ids = [ i for i in range(self.n_args) if self.arg_roles[i]=='input']
            self.n_inputs = len(self.input_dtypes)
            self.output_dtypes = [self.dtypes[arg_id] for arg_id in range(self.n_args) if self.arg_roles[arg_id] == 'output']
            self.output_shapes = [self.shapes[arg_id] for arg_id in range(self.n_args) if self.arg_roles[arg_id] == 'output']
            self.output_ids = [ i for i in range(self.n_args) if self.arg_roles[i]=='output']
            self.n_outputs = len(self.output_dtypes)
            self.scalar_dtypes = [self.dtypes[i] if self.shapes[i]==None else None for i in range(self.n_args)]
            self.call = None

            self.cl_buffers = None
    
        @staticmethod
        def read_ranges(r):
            if not isinstance(r,(np.ndarray,tuple,list,int)):
                return None
            elif isinstance(r,np.ndarray):
                return r.astype(int_type)
            elif isinstance(r,int):
                return (int_type(r),)
            else:
                return tuple(int_type(i) for i in r)
 
        def assemble_buffers(self,input_buffers=False):
            assert self.context_attached, 'openCL context is not jet attached to this instance of CLFunction. You need to call attach_context befor calling assemble_buffers.'
            ctx = self.ctx
            mf = cl.mem_flags    
            cl_args = []            
            for arg_id in range(self.n_args):
                shape = self.shapes[arg_id]
                dtype = self.dtypes[arg_id]
                arg_role = self.arg_roles[arg_id]
                if arg_role in ['input','output'] :
                    if not isinstance(input_buffers,list):
                        if shape != None:
                            _buffer = cl.Buffer(ctx,mf.READ_WRITE,size=np.prod(shape)*np.dtype(dtype).itemsize)
                            cl_args.append(_buffer)
                        else:
                            cl_args.append(None)
                    else:
                        try:
                            cl_arg.append(input_buffers.pop(0))
                        except IndexError as e:
                            log.error('There are {} input buffers needed for CLFunction {} but not enough buffers have been provided to assemble_buffers.'.format(self.n_inputs,self.name))
                            raise e
                elif arg_role =='const_input':            
                    const_input = self.const_inputs[arg_id]
                    if shape != None:
                        _buffer = cl.Buffer(context = ctx,flags =  mf.READ_ONLY | mf.COPY_HOST_PTR,size = const_input.nbytes, hostbuf=np.ascontiguousarray(const_input))
                        cl_args.append(_buffer)
                    else:
                        cl_args.append(dtype(const_input))
                elif arg_role == 'local':
                    _buffer = cl.LocalMemory(np.prod(shape)*dtype().nbytes)
                    cl_args.append(_buffer)
                else:
                    raise AssertionError('Argument role {} is not known. Known roles are (input,output,const_input,local)'.format(arg_role))
                    
                if isinstance(input_buffers,list):
                    assert len(input_buffers)==0, 'There are exactely {} input buffers needed for CLFunction {} but more buffers have been provided to assemble_buffers.'.format(self.n_inputs,self.name)
            self.cl_buffers=cl_args
            #log.info(self.cl_buffers)
            return self

        def get_function(self,input_arrays = False,output_arrays = False):
            if (not isinstance(input_arrays,(list,tuple))) and (not isinstance(output_arrays,(list,tuple))):
                fct = self.get_function_unknown_IO_arrays()
            else:
                fct = self.get_function_known_IO_arrays(input_arrays,output_arrays)
            self.call =fct
            return fct
        
        def get_function_known_IO_arrays(self,input_arrays,output_arrays):
            assert self.context_attached, 'openCL context is not jet attached to this instance of CLFunction. You need to call attach_context befor calling assemble_function.'
            if not isinstance(self.cl_buffers,list):
                self.assemble_buffers()
            kernel = self.kernel
            if isinstance(kernel, str):
                kernel = cl.Program(self.ctx,kernel).build()
                self.kernel = kernel                
            cl_fct = kernel.__getattr__(self.name)
            cl_buffers = self.cl_buffers
            #log.info(cl_buffers)
            input_ids = self.input_ids
            output_ids = self.output_ids
            
            input_buffers = [cl_buffers[i] for i in input_ids]            
            output_buffers = [cl_buffers[i] for i in output_ids]
            cl_fct.set_scalar_arg_dtypes(self.scalar_dtypes)
        
            local_range = self.local_range
            global_range = self.global_range        
            n_inputs = self.n_inputs
            n_outputs = self.n_outputs
            enqueue_copy = cl.enqueue_copy
            ascontiguousarray = np.ascontiguousarray
            queue = self.queue
            if n_outputs==1:
                output_buffer = output_buffers[0]
                output_array = output_arrays[0]
                if n_inputs == 1:
                    input_buffer = input_buffers[0]
                    input_array= input_arrays[0]
                    def fct():
                        #pushing input arrays to GPU Memory
                        #log.info('input_buffer = {}'.format(input_buffer))
                        enqueue_copy(queue,input_buffer,ascontiguousarray(input_array))            
                        #executing kernel
                        #log.info('global_range {} '.format(global_range))
                        #log.info('local_range {} '.format(local_range))
                        cl_fct(queue,global_range,local_range,*cl_buffers)
                        #extract output arrays from GPU
                        enqueue_copy(queue,output_array,output_buffer)
                else:
                    def fct():
                        #pushing input arrays to GPU Memory
                        for i in range(n_inputs):
                            enqueue_copy(queue,input_buffers[i],ascontiguousarray(input_arrays[i]))
                        #executing kernel
                        cl_fct(queue,global_range,local_range,*cl_buffers)
                        #extract output arrays from GPU
                        enqueue_copy(queue,output_array,output_buffer)
            else: 
                if n_inputs == 1:
                    input_buffer = input_buffers[0]
                    input_array= input_arrays[0]
                    def fct():
                        #pushing input arrays to GPU Memory
                        enqueue_copy(queue,input_buffer,ascontiguousarray(input_array))            
                        #executing kernel
                        cl_fct(queue,global_range,local_range,*cl_buffers)
                        #extract output arrays from GPU
                        for i in range(n_outputs):
                            enqueue_copy(queue,output_arrays[i],output_buffers[i])
                else:
                    def fct():
                        #pushing input arrays to GPU Memory
                        for i in range(n_inputs):
                            enqueue_copy(queue,input_buffers[i],ascontiguousarray(input_arrays[i]))
                        #executing kernel
                        cl_fct(queue,global_range,local_range,*cl_buffers)
                        #extract output arrays from GPU
                        for i in range(n_outputs):
                            enqueue_copy(queue,output_arrays[i],output_buffers[i])
            return fct
            
        def get_function_unknown_IO_arrays(self):
            assert self.context_attached, 'openCL context is not jet attached to this instance of CLFunction. You need to call attach_context befor calling assemble_function.'
            if not isinstance(self.cl_buffers,list):
                self.assemble_buffers()
            kernel = self.kernel
            if isinstance(kernel, str):
                kernel = cl.Program(self.ctx,kernel).build()
                self.kernel = kernel
            cl_fct = kernel.__getattr__(self.name)
            cl_buffers = self.cl_buffers
            input_ids = self.input_ids
            output_ids = self.output_ids
            output_arrays = [np.zeros(shape,dtype = dtype) for shape,dtype in zip(self.output_shapes,self.output_dtypes)]
            output_buffers = [cl_buffers[i] for i in output_ids]
            
            cl_fct.set_scalar_arg_dtypes(self.scalar_dtypes)
        
            local_range = self.local_range
            global_range = self.global_range        
            n_inputs = self.n_inputs
            n_outputs = self.n_outputs
            enqueue_copy = cl.enqueue_copy
            ascontiguousarray = np.ascontiguousarray
            queue = self.queue
            
            if n_outputs==1:
                output_arg = cl_buffers[output_ids[0]]
                output_array = output_arrays[0]
                if n_inputs == 1:
                    input_arg = cl_buffers[input_ids[0]]            
                    def fct(arg):
                        #pushing input arrays to GPU Memory
                        enqueue_copy(queue,input_arg,ascontiguousarray(arg))            
                        #executing kernel
                        cl_fct(queue,global_range,local_range,*cl_buffers)
                        #extract output arrays from GPU
                        enqueue_copy(queue,output_array,output_arg)
                        return output_array
                else:
                    def fct(*args):
                        #pushing input arrays to GPU Memory
                        for i in range(n_inputs):
                            enqueue_copy(queue,cl_buffers[input_ids[i]],ascontiguousarray(args[i]))
                        #executing kernel
                        cl_fct(queue,global_range,local_range,*cl_buffers)
                        #extract output arrays from GPU
                        enqueue_copy(queue,output_array,output_arg)
                        return output_array
            else: 
                if n_inputs == 1:
                    input_arg = cl_buffers[input_ids[0]]
                    def fct(arg):
                        #pushing input arrays to GPU Memory
                        enqueue_copy(queue,input_arg,ascontiguousarray(arg))            
                        #executing kernel
                        cl_fct(queue,global_range,local_range,*cl_buffers)
                        #extract output arrays from GPU
                        for i in range(n_outputs):
                            enqueue_copy(queue,output_arrays[i],output_buffers[i])
                        return output_arrays
                else:
                    def fct(*args):
                        #pushing input arrays to GPU Memory
                        for i in range(n_inputs):
                            enqueue_copy(queue,cl_buffers[input_ids[i]],ascontiguousarray(args[i]))
                        #executing kernel
                        cl_fct(queue,global_range,local_range,*cl_buffers)
                        #extract output arrays from GPU
                        for i in range(n_outputs):
                            enqueue_copy(queue,output_arrays[i],output_buffers[i])
                        return output_arrays
            self.call = fct
            return fct
    
    class ClProcess(ContextHandler):
        def __init__(self, process_data : dict,context_data=False):
            super().__init__()
            context_data_available = isinstance(context_data,(tuple,list))
            if context_data_available:
                super().attach_context(*context_data)
            self.dict = process_data
            self.name = process_data['name']
            #self.hash = hash((process_data['kernel'], process_data['name']))
            self.hash = hash(
                (process_data['kernel'],
                 process_data['name'],
                 tuple( (fdat['shapes'],hash_numpy_in_tuple(fdat['const_inputs'])) for fdat in process_data['functions'])
            ))
            self.kernel = process_data['kernel']
            self.functions = [OpenClPlugin.ClFunction(func_data,self.kernel,context_data) for func_data in process_data['functions']]
            self.n_functions = len(self.functions)
            self.input_dtypes = self.functions[0].input_dtypes
            self.input_shapes = self.functions[0].input_shapes
            self.n_inputs = self.functions[0].n_inputs
            self.output_dtypes = self.functions[-1].output_dtypes
            self.output_shapes = self.functions[-1].output_shapes
            self.n_outputs = self.functions[0].n_outputs
            self.run = None
        def assemble_buffers(self):
            self.functions[0].assemble_buffers()
            out_ids = self.functions[0].output_ids
            out_buffers = [self.functions[0].cl_buffers[i] for i in out_ids]
            for cl_function in self.functions[1:]:
                cl_function.assemble_buffers(input_buffers=out_buffers)
                out_ids = cl_function.output_ids
                out_buffers = [ cl_function.cl_buffers[i] for i in out_ids]
            return self
                
        def assemble(self,input_arrays=False,output_arrays=False):
            assert self.context_attached, 'openCL context is not jet attached to this instance of CLProcess. You need to call attach_context befor calling assemble.'
            if self.n_functions == 1:
                self.run = self.assemble_1fct_process(input_arrays=input_arrays,output_arrays= output_arrays)
            else:
                self.run = self.assemble_nfct_process()
            return self
        def assemble_1fct_process(self,input_arrays = False,output_arrays = False):            
            return self.functions[0].get_function(input_arrays = input_arrays,output_arrays=output_arrays)
        def asssemble_nfct_process(self):
            cl_args = [self.functions[0].assemble_buffers()]
            initial_function = self.functions[0]
            output_ids = initial_function.output_ids
            for cl_function in self.functions[1:]:
                cl_args.append(cl_function.assemble_buffers(input_buffers=cl_args[-1][output_ids]))
                output_ids = cl_functions.output_ids
            raise NotImplementedError('GPU processes in which multiple GPU functions are called in succession are not implemented jet.')

        def attach_context(self,ctx,queue):
            super().attach_context(ctx,queue)
            for function in self.functions:
                function.attach_context(ctx,queue)
            
        
    @staticmethod
    def get_number_of_gpus(allow_master_process=False):
        platform = _get_platforms(allow_master_process)
        my_gpu_devices = platform[0].get_devices(device_type=cl.device_type.GPU)
        n_gpus = len(my_gpu_devices)
        return n_gpus

    @classmethod
    def create_function(cls,kernel_data):
        fct_by_gpu = []
        for state in cls.cl_state:
            fct_per_gpu.append(cls.CLProcess(kernel_data,state['context'],state['queue']).assemble().run)    
        return fct_by_gpu
    @classmethod
    def create_process_on_all_gpus(cls, process_data:dict):
        process_by_gpu = []
        for state in cls.cl_state:
            process_by_gpu.append(cls.CLProcess(kernel_data,state['context'],state['queue']).assemble())    
        return process_by_gpu
    @classmethod
    def create_process_buffers_on_all_gpus(cls,process_data:dict):
        process_by_gpu = []
        for state in cls.cl_state:
            process_by_gpu.append(cls.ClProcess(process_data,context_data=[state['context'],state['queue']]).assemble_buffers())    
        return process_by_gpu

    

            
        



import numpy as np
import numpy.ma as mp
import logging
from itertools import repeat

log=logging.getLogger('root')

from xframe.library.mathLibrary import eval_ND_zernike_polynomials
from xframe.library.mathLibrary import bessel_jnu
from xframe.library.mathLibrary import spherical_bessel_jnu
from xframe.library.mathLibrary import polar_spherical_dft_reciprocity_relation_radial_cutoffs
from xframe.library.mathLibrary import gauss_legendre
from xframe import Multiprocessing

ht_modes = ['trapz','Zernike','midpoint','gauss',]

###legacy imports###
from xframe.library.pythonLibrary import plugVariableIntoFunction
from xframe.library.mathLibrary import spherical_bessel_jnu
import scipy.integrate as spIntegrate

def generate_weightDict(max_order, n_radial_points,reciprocity_coefficient = np.pi,dimensions=3,n_cpus=False,mode=ht_modes[0],**kwargs):
    if mode == ht_modes[0]:
        wd = generate_weightDict_trapz(max_order, n_radial_points,reciprocity_coefficient,dimensions=dimensions,n_cpus=n_cpus,**kwargs)
    elif mode == ht_modes[1]:
        wd = generate_weightDict_zernike(max_order, n_radial_points,reciprocity_coefficient,dimensions=dimensions,n_cpus=n_cpus,**kwargs)
    elif mode == ht_modes[2]:
        wd = generate_weightDict_mid(max_order, n_radial_points,reciprocity_coefficient,dimensions=dimensions,n_cpus=n_cpus,**kwargs)
    elif mode == ht_modes[3]:
        wd = generate_weightDict_gauss(max_order, n_radial_points,reciprocity_coefficient,dimensions=dimensions,n_cpus=n_cpus,**kwargs)
    else:
        raise AssertionError("Hankel transform mode {} not known. Known modes are {}".format(mode,ht_modes))
    wd['mode']=mode
    return wd
def assemble_weights(weights,pos_orders,r_max,reciprocity_coefficient=np.pi,dimensions=3,mode = ht_modes[0]):
    if mode == ht_modes[0]:
        pos_orders = np.arange(weights.shape[0])
        w = assemble_weights_trapz(weights,pos_orders,r_max,reciprocity_coefficient,dimensions=dimensions)
    elif mode == ht_modes[1]:
        w = assemble_weights_zernike(weights,pos_orders,r_max,reciprocity_coefficient,dimensions=dimensions)
    elif mode == ht_modes[2]:
        w = assemble_weights_mid(weights,pos_orders,r_max,reciprocity_coefficient,dimensions=dimensions)
    elif mode == ht_modes[3]:
        w = assemble_weights_gauss(weights,pos_orders,r_max,reciprocity_coefficient,dimensions=dimensions)
    else:
        raise AssertionError("Hankel transform mode {} not known. Known modes are {}".format(mode,ht_modes))
    w['mode']=mode
    return w

###############################################################
###    polar/spherical HT weights via Zernike expansion     ###
def generate_weightDict_zernike(max_order, n_radial_points,expansion_limit = -1,reciprocity_coefficient = np.pi,dimensions=3,n_cpus=False,approximation_type = 'trapz'):
    '''
    Calculates the weights for the Zernike version of the approximated hankel transform.
    There is no sense to chose expansion_limit higher than 2*n_radial_points-1, since this is the limit ing degree of a polynomial (i.e. also the Zernike polynomials) that could be resolved on a gauss legendre grid. In this routine we use the trapezoidal rule so it is probably advisible to stay below that limit.
    Strangly the last sentence does not seem to be right, the transform gets better for higher limits thats why, quite arbitrarily, the default below corresponds to 2*(2*n_radial_points-1)
    '''
    orders = np.arange(max_order+1)
    assert 2*n_radial_points-1>max_order,'Not enough radial points (N_r={}) to calculate weights upto order {}. max_order<2*n_radial_points-1 '.format(n_radial_points,max_order)
    
    if expansion_limit < 0:
        expansion_limit = 2*(2*n_radial_points - 1)
        log.info('Default zernike expansion limit = {}'.format(expansion_limit))
    expansion_limit = max(expansion_limit,max_order)

    worker_by_dimensions= {
        3:{
            'trapz':calc_spherical_zernike_weights,
            'midpoint':calc_spherical_zernike_weights_midpoint
        },
        2:{
            'trapz':calc_polar_zernike_weights,
            'midpoint':calc_polar_zernike_weights_midpoint
        }
    }
    worker = worker_by_dimensions[dimensions][approximation_type]
    if not isinstance(n_cpus,bool):
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,input_arrays=[orders],const_inputs=[n_radial_points,expansion_limit,reciprocity_coefficient],call_with_multiple_arguments=True,n_processes=n_cpus)
    else:
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,input_arrays=[orders],const_inputs=[n_radial_points,expansion_limit,reciprocity_coefficient],call_with_multiple_arguments=True)
    weightDict={}
    weightDict['weights'] = weights
    weightDict['posHarmOrders'] = orders
    weightDict['mode'] = 'Zerinike_{}D_{}'.format(dimensions,approximation_type)
    return weightDict

# this is equivalent to trapz for expansion_limit -> infinity -.- so no good
def calc_spherical_zernike_weights(orders,n_radial_points,expansion_limit,reciprocity_coefficient,**kwargs):
    '''
    Calculates the weights $w_{l,p,k}$ for the Zernike version of the approximated hankel transform.
    There is no sense to chose expansion_limit higher than 2*n_radial_points-1, since this is the limit ing degree of a polynomial (i.e. also the Zernike polynomials) that could be resolved on a gauss legendre grid. In this routine we use the trapezoidal rule so it is probably advisible to stay below that limit.
    $l$ indexes the order, $p$ the new_radial_coordinate and  $k$ the summation radial coordinate.
    '''    
    radial_points=n_radial_points    
    #radial_points=opt['n_radial_points']+1
    
    ps=np.arange(1,radial_points)
    n_p = radial_points-1 
    ks=np.arange(radial_points)
    #    log.info('ps={}'.format(ps/2))
    n_k = radial_points
    #weights=np.zeros((len(orders),n_p,n_k))
    #zernike_dict=eval_ND_zernike_polynomials(orders,expansion_limit,ks/radial_points,3)

    def weights_for_fixed_order(l):
        zernike_dict=eval_ND_zernike_polynomials(np.array([l]),expansion_limit,ps/radial_points,3)
        #log.info('finished_zernike dict')
        s=np.arange(l,expansion_limit+1,2)
        len_s=len(s)
        prefactor=(-1)**( (s-l)/2 )*(2*s+3)
        jp=spherical_bessel_jnu(np.repeat((s+1)[:,None],radial_points-1,axis=1),ks[1:]*reciprocity_coefficient)
        #log.info('jp shape =  {}'.format(jp.shape))
        Zk=zernike_dict[l]
        #log.info('Zk shape = {}'.format(Zk.shape))
        # log.info('prefactor = \n {}'.format(prefactor))
        #log.info('Zk = \n {}'.format(Zk))
        #log.info('making summands for order = {}'.format(l))
        summands = np.zeros((len_s,n_p,n_k))
        summands[:,:,1:] = prefactor[:,None,None]*Zk[:,:,None]*jp[:,None,:]
        if l == 0:
            summands[0,:,0] = reciprocity_coefficient ## this might be trouble ? previousely it was pi or 1/2
        weights=np.sum(summands,axis=0)
        #log.info('w sum = {}'.format(weights.sum()))#
        return weights
    weights=np.array(tuple(map(weights_for_fixed_order,orders)))
    c_kp=np.zeros((n_p,n_k))
    c_kp[:,1:]=np.square(ps)[:,None]/(ks[None,1:])
    c_kp[:,0]=np.square(ps)
    weights=weights*c_kp[None,:,:]
    return weights

# this is equivalent to trapz for expansion_limit -> infinity -.- so no good
def calc_polar_zernike_weights(orders,n_radial_points,expansion_limit,reciprocity_coefficient,**kwargs):
    '''
     Calculates the weights $w_{l,p,k}$ for the Zernike version of the approximated hankel transform.
     There is no sense to chose expansion_limit higher than 2*n_radial_points-1, since this is the limit ing degree of a polynomial (i.e. also the Zernike polynomials) that could be resolved on a gauss legendre grid. In this routine we use the trapezoidal rule so it is probably advisible to stay below that limit.
     $l$ indexes the order, $p$ the new_radial_coordinate and  $k$ the summation radial coordinate.
    '''    

    radial_points=n_radial_points    
    #radial_points=opt['n_radial_points']+1
    
    ps=np.arange(1,radial_points)
    n_p = radial_points-1 
    ks=np.arange(radial_points)
    #    log.info('ps={}'.format(ps/2))
    n_k = radial_points
    #weights=np.zeros((len(orders),n_p,n_k))
    #zernike_dict=eval_ND_zernike_polynomials(orders,expansion_limit,ks/radial_points,3)

    def weights_for_fixed_order(m):
        zernike_dict=eval_ND_zernike_polynomials(np.array([m]),expansion_limit,ps/radial_points,2)
        #log.info('finished_zernike dict')
        s=np.arange(m,expansion_limit+1,2)
        len_s=len(s)
        prefactor=(-1)**( (s-m)/2 )*(2*s+2)
        Jk = bessel_jnu(np.repeat((s+1)[:,None],radial_points-1,axis=1),ks[1:]*reciprocity_coefficient)
        #log.info('jp shape =  {}'.format(jp.shape))
        Zp=zernike_dict[m]
        #log.info('Zk shape = {}'.format(Zk.shape))
        # log.info('prefactor = \n {}'.format(prefactor))
        #log.info('Zk = \n {}'.format(Zk))
        log.info('making summands for order = {}'.format(m))
        summands = np.zeros((len_s,n_p,n_k))
        summands[:,:,1:] = prefactor[:,None,None]*Zp[:,:,None]*Jk[:,None,:]
        if m == 0:
            summands[0,:,0] = reciprocity_coefficient
        weights=np.sum(summands,axis=0)
        #log.info('w sum = {}'.format(weights.sum()))#
        return weights
    weights=np.array(tuple(map(weights_for_fixed_order,orders)))
    c_kp=np.zeros((n_p,n_k))
    c_kp[:,1:]=ps[:,None]/(ks[None,1:])
    c_kp[:,0]=ps
    weights=weights*c_kp[None,:,:]
    return weights


# this is equivalent to midpoint for expansion_limit -> infinity -.- so no good
def calc_spherical_zernike_weights_midpoint(orders,n_radial_points,expansion_limit,reciprocity_coefficient,**kwargs):
    '''
    Calculates the weights $w_{l,p,k}$ for the Zernike version of the approximated hankel transform.
    There is no sense to chose expansion_limit higher than 2*n_radial_points-1, since this is the limit ing degree of a polynomial (i.e. also the Zernike polynomials) that could be resolved on a gauss legendre grid. In this routine we use the trapezoidal rule so it is probably advisible to stay below that limit.
    $l$ indexes the order, $p$ the new_radial_coordinate and  $k$ the summation radial coordinate.
    '''    
    radial_points=n_radial_points    
    #radial_points=opt['n_radial_points']+1
    
    ps=np.arange(radial_points)+1/2
    n_p = radial_points
    ks=np.arange(radial_points)+1/2
    #    log.info('ps={}'.format(ps/2))
    n_k = radial_points
    #weights=np.zeros((len(orders),n_p,n_k))
    #zernike_dict=eval_ND_zernike_polynomials(orders,expansion_limit,ks/radial_points,3)

    def weights_for_fixed_order(l):
        zernike_dict=eval_ND_zernike_polynomials(np.array([l]),expansion_limit,ps/radial_points,3)
        #log.info('finished_zernike dict')
        s=np.arange(l,expansion_limit+1,2)
        len_s=len(s)
        prefactor=(-1)**( (s-l)/2 )*(2*s+3)
        jp=spherical_bessel_jnu(np.repeat((s+1)[:,None],radial_points,axis=1),ks*reciprocity_coefficient)
        #log.info('jp shape =  {}'.format(jp.shape))
        Zk=zernike_dict[l]
        #log.info('Zk shape = {}'.format(Zk.shape))
        # log.info('prefactor = \n {}'.format(prefactor))
        #log.info('Zk = \n {}'.format(Zk))
        #log.info('making summands for order = {}'.format(l))
        summands = np.zeros((len_s,n_p,n_k))
        summands[:,:,:] = prefactor[:,None,None]*Zk[:,:,None]*jp[:,None,:]
        #if l == 0:
        #    summands[0,:,0] = reciprocity_coefficient ## this might be trouble ? previousely it was pi or 1/2
        weights=np.sum(summands,axis=0)
        #log.info('w sum = {}'.format(weights.sum()))#
        return weights
    weights=np.array(tuple(map(weights_for_fixed_order,orders)))
    c_kp=np.zeros((n_p,n_k))
    c_kp[:,:]=np.square(ps)[:,None]/(ks[None,:])
    #c_kp[:,0]=np.square(ps)
    weights=weights*c_kp[None,:,:]
    return weights

# this is equivalent to midpoint for expansion_limit -> infinity -.- so no good
def calc_polar_zernike_weights_midpoint(orders,n_radial_points,expansion_limit,reciprocity_coefficient,**kwargs):
    '''
     Calculates the weights $w_{l,p,k}$ for the Zernike version of the approximated hankel transform.
     There is no sense to chose expansion_limit higher than 2*n_radial_points-1, since this is the limit ing degree of a polynomial (i.e. also the Zernike polynomials) that could be resolved on a gauss legendre grid. In this routine we use the trapezoidal rule so it is probably advisible to stay below that limit.
     $l$ indexes the order, $p$ the new_radial_coordinate and  $k$ the summation radial coordinate.
    '''    

    radial_points=n_radial_points    
    #radial_points=opt['n_radial_points']+1
    
    ps=np.arange(radial_points)+1/2
    n_p = radial_points
    ks=np.arange(radial_points)+1/2
    #    log.info('ps={}'.format(ps/2))
    n_k = radial_points
    #weights=np.zeros((len(orders),n_p,n_k))
    #zernike_dict=eval_ND_zernike_polynomials(orders,expansion_limit,ks/radial_points,3)

    def weights_for_fixed_order(m):
        zernike_dict=eval_ND_zernike_polynomials(np.array([m]),expansion_limit,ps/radial_points,2)
        #log.info('finished_zernike dict')
        s=np.arange(m,expansion_limit+1,2)
        len_s=len(s)
        prefactor=(-1)**( (s-m)/2 )*(2*s+2)
        Jk = bessel_jnu(np.repeat((s+1)[:,None],radial_points,axis=1),ks*reciprocity_coefficient)
        #log.info('jp shape =  {}'.format(jp.shape))
        Zp=zernike_dict[m]
        #log.info('Zk shape = {}'.format(Zk.shape))
        # log.info('prefactor = \n {}'.format(prefactor))
        #log.info('Zk = \n {}'.format(Zk))
        log.info('making summands for order = {}'.format(m))
        summands = np.zeros((len_s,n_p,n_k))
        summands[:,:,:] = prefactor[:,None,None]*Zp[:,:,None]*Jk[:,None,:]
        #if m == 0:
        #    summands[0,:,0] = reciprocity_coefficient
        weights=np.sum(summands,axis=0)
        #log.info('w sum = {}'.format(weights.sum()))#
        return weights
    weights=np.array(tuple(map(weights_for_fixed_order,orders)))
    c_kp=np.zeros((n_p,n_k))
    c_kp[:,:]=ps[:,None]/(ks[None,:])
    #c_kp[:,0]=ps
    weights=weights*c_kp[None,:,:]
    return weights

def assemble_weights_zernike(weights,orders,r_max,reciprocity_coefficient,dimensions=3):
    '''
    Generates weights for the forward and inverse transform by multiplieing with the propper constants.
    And changes the Array order of the weights from (order,summation_radial_corrdinate,new_radial_coordinate) to 
    (summation_radial_coordinate,new_radial_coordinate,order)
    '''
    n_radial_points = weights.shape[-1]
    q_max=polar_spherical_dft_reciprocity_relation_radial_cutoffs(r_max,n_radial_points,reciprocity_coefficient=reciprocity_coefficient)

    if dimensions == 3:
        #forward_prefactor = (-1.j)**(orders[None,None,:])*(np.pi**2/q_max**3)*np.sqrt(2/np.pi)
        forward_prefactor = (-1.j)**(orders[None,None,:])*(r_max/n_radial_points)**3*np.sqrt(2/np.pi**3)
        #inverse_prefactor = (1.j)**(orders[None,None,:])*(np.pi**2/r_max**3)*np.sqrt(2/np.pi)
        inverse_prefactor = (1.j)**(orders[None,None,:])*(q_max/n_radial_points)**3*np.sqrt(2/np.pi**3)
    elif dimensions == 2:
        all_orders = np.concatenate((orders,orders[:0:-1]))
        #forward_prefactor = (-1.j)**(all_orders[None,None,:])*(np.pi/q_max**2)
        forward_prefactor = (-1.j)**(all_orders[None,None,:])*(r_max/n_radial_points)**2/np.pi
        #inverse_prefactor = (1.j)**(all_orders[None,None,:])*(np.pi/r_max**2)
        inverse_prefactor = (1.j)**(all_orders[None,None,:])*(q_max/n_radial_points)**2/np.pi
        # weights for negative orders are given by $w_{-mpk}=(-1)^m $w_{mpk}$ due to $J_{-m}(x)=(-1)^m J_m(x)$.
        weights = np.concatenate((weights,(-1)**orders[:0:-1,None,None]*weights[:0:-1]),axis = 0)   
        
    weights=np.moveaxis(weights,0,2)
    forward_weights=weights*forward_prefactor
    inverse_weights=weights*inverse_prefactor
    #log.info('weights shape = {}'.format(forward_weights.shape))
    return {'forward':forward_weights,'inverse':inverse_weights}


#############################################################
###    polar/spherical HT weights via trapezoidal rule    ###
def generate_weightDict_trapz(max_order, n_radial_points,reciprocity_coefficient = np.pi,dimensions=3,n_cpus=False):
    '''    
    '''
    orders = np.arange(max_order+1)
    #assert 2*n_radial_points-1>max_order,'Not enough radial points (N_r={}) to calculate weights upto order {}. max_order<2*n_radial_points-1 '.format(n_radial_points,max_order)
    

    worker_by_dimensions= {2:calc_polar_trapz_weights,3:calc_spherical_trapz_weights}
    worker = worker_by_dimensions[dimensions]
    if not isinstance(n_cpus,bool):
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,input_arrays=[orders],const_inputs=[n_radial_points,reciprocity_coefficient],call_with_multiple_arguments=True,n_processes=n_cpus)
    else:
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,input_arrays=[orders],const_inputs=[n_radial_points,reciprocity_coefficient],call_with_multiple_arguments=True)
    weightDict={}
    weightDict['weights'] = weights
    weightDict['posHarmOrders'] = orders
    weightDict['mode'] = 'trapz'
    weightDict['dimension'] = dimensions
    return weightDict

def calc_spherical_trapz_weights(orders,n_radial_points,reciprocity_coefficient,**kwargs):
    '''
    Generates weights for trapezooidal rule
    '''
    N = n_radial_points
    ps=np.arange(1,N)
    ks=np.arange(N)

    ls = orders
    jmpk = spherical_bessel_jnu(np.repeat(np.repeat(ls[:,None,None],N-1,axis=1),N,axis=2),ks[None,:]*ps[:,None]*reciprocity_coefficient/N)
    weights = ps[None,:,None]**2*jmpk
    return weights

def calc_polar_trapz_weights(orders,n_radial_points,reciprocity_coefficient,**kwargs):
    '''
    Generates weights for trapezoidal rule.
    '''
    N = n_radial_points
    ps=np.arange(1,N)
    ks=np.arange(N)
    ms = orders
    Jmpk = bessel_jnu(np.repeat(np.repeat(ms[:,None,None],N-1,axis=1),N,axis=2),ks[None,:]*ps[:,None]*reciprocity_coefficient/N)
    zero_order_mask = (ms == 0)
    #Jmpk[zero_order_mask]*=np.sqrt(2)
    weights = ps[None,:,None]*Jmpk
    return weights

def assemble_weights_trapz(weights,orders,r_max,reciprocity_coefficient,dimensions=3):
    '''
    Generates weights for the forward and inverse transform by multiplieing with the propper constants.
    And changes the Array order of the weights from (order,summation_radial_corrdinate,new_radial_coordinate) to 
    (summation_radial_coordinate,new_radial_coordinate,order)
    '''
    n_radial_points = weights.shape[-1]
    q_max=polar_spherical_dft_reciprocity_relation_radial_cutoffs(r_max,n_radial_points,reciprocity_coefficient = reciprocity_coefficient)
    #q_max=polar_spherical_dft_reciprocity_relation_radial_cutoffs_new(r_max,n_radial_points,scale = pi_in_q)
    
    if dimensions == 2:
        all_orders = np.concatenate((orders,-orders[:0:-1]))
        #log.info(all_orders)
        forward_prefactor = (-1.j)**(all_orders[None,None,:])*(r_max/n_radial_points)**2#*np.sqrt(2)
        inverse_prefactor = (1.j)**(all_orders[None,None,:])*(q_max/n_radial_points)**2#*np.sqrt(2)
        # weights for negative orders are given by $w_{-mpk}=(-1)^m $w_{mpk}$ due to $J_{-m}(x)=(-1)^m J_m(x)$.
        #log.info('wights.shape={},orders = {}'.format(weights.shape,orders))
        weights = np.concatenate((weights,(-1)**orders[:0:-1,None,None]*weights[:0:-1]),axis = 0)
    elif dimensions == 3:
        forward_prefactor = (-1.j)**(orders[None,None,:])*(r_max/n_radial_points)**3*np.sqrt(2/np.pi)
        inverse_prefactor = (1.j)**(orders[None,None,:])*(q_max/n_radial_points)**3*np.sqrt(2/np.pi)
        
    weights=np.moveaxis(weights,0,2)
    forward_weights=weights*forward_prefactor
    inverse_weights=weights*inverse_prefactor
    #log.info('weights shape = {}'.format(forward_weights.shape))
    return {'forward':forward_weights,'inverse':inverse_weights}


##########################################################
###    polar/spherical HT weights via midpoint rule    ###
def generate_weightDict_mid(max_order, n_radial_points,reciprocity_coefficient = np.pi,dimensions=3,n_cpus=False):
    '''    
    '''
    orders = np.arange(max_order+1)
    #assert 2*n_radial_points-1>max_order,'Not enough radial points (N_r={}) to calculate weights upto order {}. max_order<2*n_radial_points-1 '.format(n_radial_points,max_order)
    
    worker_by_dimensions= {2:calc_polar_mid_weights,3:calc_spherical_mid_weights}
    worker = worker_by_dimensions[dimensions]
    if not isinstance(n_cpus,bool):
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,input_arrays=[orders],const_inputs=[n_radial_points,reciprocity_coefficient],call_with_multiple_arguments=True,n_processes=n_cpus)
    else:
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,input_arrays=[orders],const_inputs=[n_radial_points,reciprocity_coefficient],call_with_multiple_arguments=True)
    weightDict={}
    weightDict['weights'] = weights
    weightDict['posHarmOrders'] = orders
    weightDict['mode'] = 'trapz'
    weightDict['dimension'] = dimensions
    return weightDict

def calc_spherical_mid_weights(orders,n_radial_points,reciprocity_coefficient,**kwargs):
    '''
    Generates weights for trapezooidal rule
    '''
    N = n_radial_points
    ps=np.arange(N)+1/2
    ks=np.arange(N)+1/2

    ls = orders
    jmpk = spherical_bessel_jnu(np.repeat(np.repeat(ls[:,None,None],N,axis=1),N,axis=2),ks[None,:]*ps[:,None]*reciprocity_coefficient/N)
    weights = ps[None,:,None]**2*jmpk
    return weights

def calc_polar_mid_weights(orders,n_radial_points,reciprocity_coefficient,**kwargs):
    '''
    Generates weights for trapezoidal rule.
    '''
    N = n_radial_points
    ps=np.arange(N)+1/2
    ks=np.arange(N)+1/2
    ms = orders
    Jmpk = bessel_jnu(np.repeat(np.repeat(ms[:,None,None],N,axis=1),N,axis=2),ks[None,:]*ps[:,None]*reciprocity_coefficient/N)
    #zero_order_mask = (ms == 0)
    #Jmpk[zero_order_mask]*=np.sqrt(2)
    weights = ps[None,:,None]*Jmpk
    return weights

def assemble_weights_mid(weights,orders,r_max,reciprocity_coefficient,dimensions=3):
    '''
    Generates weights for the forward and inverse transform by multiplieing with the propper constants.
    And changes the Array order of the weights from (order,summation_radial_corrdinate,new_radial_coordinate) to 
    (summation_radial_coordinate,new_radial_coordinate,order)
    '''
    n_radial_points = weights.shape[-1]
    q_max=polar_spherical_dft_reciprocity_relation_radial_cutoffs(r_max,n_radial_points,reciprocity_coefficient = reciprocity_coefficient)
    #q_max=polar_spherical_dft_reciprocity_relation_radial_cutoffs_new(r_max,n_radial_points,scale = pi_in_q)
    
    if dimensions == 2:
        all_orders = np.concatenate((orders,-orders[:0:-1]))
        #log.info(all_orders)
        forward_prefactor = (-1.j)**(all_orders[None,None,:])*(r_max/n_radial_points)**2#*np.sqrt(2)
        inverse_prefactor = (1.j)**(all_orders[None,None,:])*(q_max/n_radial_points)**2#*np.sqrt(2)
        # weights for negative orders are given by $w_{-mpk}=(-1)^m $w_{mpk}$ due to $J_{-m}(x)=(-1)^m J_m(x)$.
        #log.info('wights.shape={},orders = {}'.format(weights.shape,orders))
        weights = np.concatenate((weights,(-1)**orders[:0:-1,None,None]*weights[:0:-1]),axis = 0)
    elif dimensions == 3:
        forward_prefactor = (-1.j)**(orders[None,None,:])*(r_max/n_radial_points)**3*np.sqrt(2/np.pi)
        inverse_prefactor = (1.j)**(orders[None,None,:])*(q_max/n_radial_points)**3*np.sqrt(2/np.pi)
        
    weights=np.moveaxis(weights,0,2)
    forward_weights=weights*forward_prefactor
    inverse_weights=weights*inverse_prefactor
    #log.info('weights shape = {}'.format(forward_weights.shape))
    return {'forward':forward_weights,'inverse':inverse_weights}


######################################################################
###    polar/spherical HT weights via Gauss Legendre quadrature    ###
def generate_weightDict_gauss(max_order, n_radial_points,reciprocity_coefficient = np.pi,dimensions=3,n_cpus=False):
    '''    
    '''
    orders = np.arange(max_order+1)
    #assert 2*n_radial_points-1>max_order,'Not enough radial points (N_r={}) to calculate weights upto order {}. max_order<2*n_radial_points-1 '.format(n_radial_points,max_order)
    

    worker_by_dimensions= {2:calc_polar_gauss_weights,3:calc_spherical_gauss_weights}
    worker = worker_by_dimensions[dimensions]
    if not isinstance(n_cpus,bool):
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,input_arrays=[orders],const_inputs=[n_radial_points,reciprocity_coefficient],call_with_multiple_arguments=True,n_processes=n_cpus)
    else:
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,input_arrays=[orders],const_inputs=[n_radial_points,reciprocity_coefficient],call_with_multiple_arguments=True)
    weightDict={}
    weightDict['weights'] = weights
    weightDict['posHarmOrders'] = orders
    weightDict['mode'] = 'trapz'
    weightDict['dimension'] = dimensions
    return weightDict

def calc_spherical_gauss_weights(orders,n_radial_points,reciprocity_coefficient,**kwargs):
    '''
    Generates weights for gauss legendre quadrature.
    '''
    N = n_radial_points
    xi,wgauss = gauss_legendre(N)
    
    ps=xi+1
    ks=xi+1

    ls = orders
    jmpk = spherical_bessel_jnu(np.repeat(np.repeat(ls[:,None,None],N,axis=1),N,axis=2),ks[None,:]*ps[:,None]*reciprocity_coefficient*N/4)
    weights = ps[None,:,None]**2*jmpk*wgauss[None,:,None]
    return weights

def calc_polar_gauss_weights(orders,n_radial_points,reciprocity_coefficient,**kwargs):
    '''
    Generates weights for gauss legendre quadrature.
    '''

    N = n_radial_points
    xi,wgauss = gauss_legendre(N)
    
    ps=xi+1
    ks=xi+1
    ms = orders
    Jmpk = bessel_jnu(np.repeat(np.repeat(ms[:,None,None],N,axis=1),N,axis=2),ks[None,:]*ps[:,None]*reciprocity_coefficient*N/4)
    #zero_order_mask = (ms == 0)
    #Jmpk[zero_order_mask]*=np.sqrt(2)
    weights = ps[None,:,None]*Jmpk*wgauss[None,:,None]
    return weights

def assemble_weights_gauss(weights,orders,r_max,reciprocity_coefficient,dimensions=3):
    '''
    Generates weights for the forward and inverse transform by multiplieing with the propper constants.
    And changes the Array order of the weights from (order,summation_radial_corrdinate,new_radial_coordinate) to 
    (summation_radial_coordinate,new_radial_coordinate,order)
    '''
    n_radial_points = weights.shape[-1]
    q_max=polar_spherical_dft_reciprocity_relation_radial_cutoffs(r_max,n_radial_points,reciprocity_coefficient = reciprocity_coefficient)
    #q_max=polar_spherical_dft_reciprocity_relation_radial_cutoffs_new(r_max,n_radial_points,scale = pi_in_q)
    
    if dimensions == 2:
        all_orders = np.concatenate((orders,-orders[:0:-1]))
        #log.info(all_orders)
        forward_prefactor = (-1.j)**(all_orders[None,None,:])*(r_max/2)**2#*np.sqrt(2)
        inverse_prefactor = (1.j)**(all_orders[None,None,:])*(q_max/2)**2#*np.sqrt(2)
        # weights for negative orders are given by $w_{-mpk}=(-1)^m $w_{mpk}$ due to $J_{-m}(x)=(-1)^m J_m(x)$.
        #log.info('wights.shape={},orders = {}'.format(weights.shape,orders))
        weights = np.concatenate((weights,(-1)**orders[:0:-1,None,None]*weights[:0:-1]),axis = 0)
    elif dimensions == 3:
        forward_prefactor = (-1.j)**(orders[None,None,:])*(r_max/2)**3*np.sqrt(2/np.pi)
        inverse_prefactor = (1.j)**(orders[None,None,:])*(q_max/2)**3*np.sqrt(2/np.pi)
        
    weights=np.moveaxis(weights,0,2)
    forward_weights=weights*forward_prefactor
    inverse_weights=weights*inverse_prefactor
    #log.info('weights shape = {}'.format(forward_weights.shape))
    return {'forward':forward_weights,'inverse':inverse_weights}


############################################################
###     create polar/spherical HT for given weights      ###
def generate_ht(weights,used_orders,r_max,reciprocity_coefficient=np.pi,dimensions=3,use_gpu=False,mode = ht_modes[0]):
    '''
    Routine that selects the construciton routine for 2D/3D Hankel transforms via zernike expansion. 
    '''
    w = assemble_weights(weights,used_orders,r_max,reciprocity_coefficient,dimensions=dimensions,mode=mode)

    spherical_gpu_version = (dimensions == 3) and use_gpu
    spherical_version = (dimensions == 3) and (not use_gpu)
    polar_version = (dimensions == 2)  and (not use_gpu)
    polar_gpu_version = (dimensions == 2) and use_gpu
    
    if spherical_gpu_version:
        zht,izht = generate_spherical_ht_gpu(w,used_orders)
    elif spherical_version:
        zht,izht = generate_spherical_ht(w,used_orders)
    elif polar_version:
        zht,izht = generate_polar_ht(w,used_orders)
    elif polar_gpu_version:
        zht,izht = generate_polar_ht_gpu(w,used_orders)
    return zht,izht

def generate_polar_zernike_ht(w,used_orders):
    '''
    Generates polar hankel transform using Zernike weights. $HT_m(f_m) = \sum_k f_m(k)*w_kpm$ for $m\geq 0$.
    '''
    forward_weights= w['forward']
    inverse_weights= w['inverse']
    #log.info('weight shape = {}'.format(forward_weights.shape))
    n_orders = int((forward_weights.shape[-1]+1)//2)
    n_radial_steps = forward_weights.shape[1]
    all_orders = np.concatenate((np.arange(n_orders),np.arange(n_orders)[:0:-1]))
    
    unused_order_mask = ~np.in1d(all_orders,used_orders)
    
    full_shape = (n_radial_steps,len(all_orders))
    out_forward = np.zeros(full_shape,dtype = complex)
    out_inverse = np.zeros(full_shape,dtype = complex)
    def zht(harmonic_coeff):
        #log.info('harmonic shape = {}'.format(harmonic_coeff.shape))
        #log.info('out forward = {}'.format(out_forward.shape))
        reciprocal_harmonic_coeff = np.sum(forward_weights*harmonic_coeff[1:,None,:],axis = 0,out=out_forward)
        reciprocal_harmonic_coeff[:,unused_order_mask]=0
        return reciprocal_harmonic_coeff
    def izht(reciprocal_coeff):
        harmonic_coeff = np.sum(inverse_weights*reciprocal_coeff[1:,None,:],axis = 0,out=out_inverse)
        harmonic_coeff[:,unused_order_mask]=0
        return harmonic_coeff
    return zht,izht

def generate_spherical_zernike_ht(w,l_orders):
    forward_weights= w['forward']
    inverse_weights= w['inverse']

    l_max=l_orders.max()
    m_orders=np.concatenate((np.arange(l_max+1,dtype = int),-np.arange(l_max,0,-1,dtype = int)))
    def zht(harmonic_coeff):
        return tuple(np.sum(forward_weights[:,:,np.abs(m):]*harmonic_coeff[m][1:,None,:l_max - np.abs(m)+1],axis=0) for m in m_orders) 
    def izht(reciprocal_coeff):
        return tuple(np.sum(inverse_weights[:,:,np.abs(m):]*reciprocal_coeff[m][1:,None,:l_max - np.abs(m)+1],axis=0) for m in m_orders)
    return zht,izht


def generate_polar_ht(w,used_orders):
    '''
    Generates polar hankel transform using Zernike weights. $HT_m(f_m) = \sum_k f_m(k)*w_kpm$ for $m\geq 0$.
    '''
    forward_weights= w['forward']
    inverse_weights= w['inverse']
    #log.info('weight shape = {}'.format(forward_weights.shape))
    n_orders = int((forward_weights.shape[-1]+1)//2)
    n_radial_steps = forward_weights.shape[1]
    all_orders = np.concatenate((np.arange(n_orders),np.arange(n_orders)[:0:-1]))
    
    unused_order_mask = ~np.in1d(all_orders,used_orders)
    
    full_shape = (n_radial_steps,len(all_orders))
    out_forward = np.zeros(full_shape,dtype = complex)
    out_inverse = np.zeros(full_shape,dtype = complex)
    if w['mode'] in ht_modes[:2]:
        def zht(harmonic_coeff):
            #log.info('harmonic shape = {}'.format(harmonic_coeff.shape))
            #log.info('out forward = {}'.format(out_forward.shape))
            reciprocal_harmonic_coeff = np.sum(forward_weights*harmonic_coeff[1:,None,:],axis = 0,out=out_forward)
            reciprocal_harmonic_coeff[:,unused_order_mask]=0
            return reciprocal_harmonic_coeff
        def izht(reciprocal_coeff):
            harmonic_coeff = np.sum(inverse_weights*reciprocal_coeff[1:,None,:],axis = 0,out=out_inverse)
            harmonic_coeff[:,unused_order_mask]=0
            return harmonic_coeff
    else:
        def zht(harmonic_coeff):
            #log.info('harmonic shape = {}'.format(harmonic_coeff.shape))
            #log.info('out forward = {}'.format(out_forward.shape))
            reciprocal_harmonic_coeff = np.sum(forward_weights*harmonic_coeff[:,None,:],axis = 0,out=out_forward)
            reciprocal_harmonic_coeff[:,unused_order_mask]=0
            return reciprocal_harmonic_coeff
        def izht(reciprocal_coeff):
            harmonic_coeff = np.sum(inverse_weights*reciprocal_coeff[:,None,:],axis = 0,out=out_inverse)
            harmonic_coeff[:,unused_order_mask]=0
            return harmonic_coeff
    return zht,izht

def generate_spherical_ht(w,l_orders):
    forward_weights= w['forward']
    inverse_weights= w['inverse']

    l_max=l_orders.max()
    m_orders=np.concatenate((np.arange(l_max+1,dtype = int),-np.arange(l_max,0,-1,dtype = int)))
    if w['mode'] in ht_modes[:2]:
        def zht(harmonic_coeff):
            return tuple(np.sum(forward_weights[:,:,np.abs(m):]*harmonic_coeff[m][1:,None,:l_max - np.abs(m)+1],axis=0) for m in m_orders) 
        def izht(reciprocal_coeff):
            return tuple(np.sum(inverse_weights[:,:,np.abs(m):]*reciprocal_coeff[m][1:,None,:l_max - np.abs(m)+1],axis=0) for m in m_orders)
    else:
        def zht(harmonic_coeff):
            return tuple(np.sum(forward_weights[:,:,np.abs(m):]*harmonic_coeff[m][:,None,:l_max - np.abs(m)+1],axis=0) for m in m_orders) 
        def izht(reciprocal_coeff):
            return tuple(np.sum(inverse_weights[:,:,np.abs(m):]*reciprocal_coeff[m][:,None,:l_max - np.abs(m)+1],axis=0) for m in m_orders)
    return zht,izht

def generate_spherical_ht_gpu(w,l_orders):
    forward_weights= w['forward']
    inverse_weights= w['inverse']
    n_radial_points = forward_weights.shape[1]

    l_max=l_orders.max()

    nq = n_radial_points
    nl = len(l_orders)
    nlm = l_max*(l_max+2)+1
    if w['mode'] in ht_modes[:2]:        
        kernel_str = """
        __kernel void
        apply_weights(__global double2* out, 
        __global double2* w, 
        __global double2* rho, 
        long nq,long nlm, long nl)
        {
  
        long i = get_global_id(0); 
        long j = get_global_id(1);
        long l = (long) sqrt((double)j);

 
        // value stores the element that is 
        // computed by the thread
        double2 value = 0;
        // wlm is of shape (sum_q,nq,m) where sum_q = nq-1
        for (int q = 0; q < nq-1; ++q)
        {
        double2 wqql = w[q*nq*nl + i*nl + l];
        double2 rqlm = rho[(q+1)*nlm + j];
        value.x += wqql.x * rqlm.x - wqql.y * rqlm.y;
        value.y += wqql.x * rqlm.y + wqql.y * rqlm.x;
        }
    
        // Write the matrix to device memory each 
        // thread writes one element
        out[i * nlm + j] = value;//w[nq*nl+i*nl + l];
        }
        """
    else:
               kernel_str = """
        __kernel void
        apply_weights(__global double2* out, 
        __global double2* w, 
        __global double2* rho, 
        long nq,long nlm, long nl)
        {
  
        long i = get_global_id(0); 
        long j = get_global_id(1);
        long l = (long) sqrt((double)j);

 
        // value stores the element that is 
        // computed by the thread
        double2 value = 0;
        // wlm is of shape (sum_q,nq,m) where sum_q = nq-1
        for (int q = 0; q < nq; ++q)
        {
        double2 wqql = w[q*nq*nl + i*nl + l];
        double2 rqlm = rho[q*nlm + j];
        value.x += wqql.x * rqlm.x - wqql.y * rqlm.y;
        value.y += wqql.x * rqlm.y + wqql.y * rqlm.x;
        }
    
        // Write the matrix to device memory each 
        // thread writes one element
        out[i * nlm + j] = value;//w[nq*nl+i*nl + l];
        }
        """

    kernel_dict_forward={
            'kernel': kernel_str,
            'name': 'forward_hankel',
            'functions': ({
                'name': 'apply_weights',
                'dtypes' : (complex,complex,complex,np.int64,np.int64,np.int64),
                'shapes' : ((nq,nlm),forward_weights.shape,(nq,nlm),None,None,None),
                'arg_roles' : ('output','const_input','input','const_input','const_input','const_input'),
                'const_inputs' : (None,forward_weights,None,np.int64(nq),np.int64(nlm),np.int64(nl)),
                'global_range' : (nq,nlm),
                'local_range' : None
            },)
        }
    
    kernel_dict_inverse={
            'kernel': kernel_str,
            'name': 'inverse_hankel',
            'functions': ({
                'name': 'apply_weights',
                'dtypes' : (complex,complex,complex,np.int64,np.int64,np.int64),
                'shapes' : ((nq,nlm),inverse_weights.shape,(nq,nlm),None,None,None),
                'arg_roles' : ('output','const_input','input','const_input','const_input','const_input'),
                'const_inputs' : (None,inverse_weights,None,np.int64(nq),np.int64(nlm),np.int64(nl)),
                'global_range' : (nq,nlm),
                'local_range' : None
            },)
        }
    
    forward_gpu_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_forward)
    inverse_gpu_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_inverse)

    zernike_hankel_transform = Multiprocessing.comm_module.add_gpu_process(forward_gpu_process)
    inverse_zernike_hankel_transform = Multiprocessing.comm_module.add_gpu_process(inverse_gpu_process)
    return zernike_hankel_transform,inverse_zernike_hankel_transform

def generate_polar_ht_gpu(w,used_orders):
    forward_weights= w['forward']
    inverse_weights= w['inverse']
    n_radial_points = forward_weights.shape[1]

    m_max=used_orders.max()

    nq = n_radial_points
    nm = m_max*2+1

    if w['mode'] in ht_modes[:2]:
        kernel_str = """
        __kernel void
        apply_weights(__global double2* out, 
        __global double2* w, 
        __global double2* rho, 
        long nq,long nm)
        {
  
        long i = get_global_id(0); 
        long j = get_global_id(1);
 
        // value stores the element that is 
        // computed by the thread
        double2 value = 0;
        // wlm is of shape (sum_q,nq,m) where sum_q = nq-1
        for (int q = 0; q < nq-1; ++q)
        {
        double2 wqqm = w[q*nq*nm + i*nm + j];
        double2 rqm = rho[(q+1)*nm + j];
        value.x += wqqm.x * rqm.x - wqqm.y * rqm.y;
        value.y += wqqm.x * rqm.y + wqqm.y * rqm.x;
        }
    
        // Write the matrix to device memory each 
        // thread writes one element
        out[i * nm + j] = value;//w[nq*nl+i*nl + l];
        }
        """
    else:
        kernel_str = """
        __kernel void
        apply_weights(__global double2* out, 
        __global double2* w, 
        __global double2* rho, 
        long nq,long nm)
        {
  
        long i = get_global_id(0); 
        long j = get_global_id(1);
 
        // value stores the element that is 
        // computed by the thread
        double2 value = 0;
        // wlm is of shape (sum_q,nq,m) where sum_q = nq-1
        for (int q = 0; q < nq; ++q)
        {
        double2 wqqm = w[q*nq*nm + i*nm + j];
        double2 rqm = rho[q*nm + j];
        value.x += wqqm.x * rqm.x - wqqm.y * rqm.y;
        value.y += wqqm.x * rqm.y + wqqm.y * rqm.x;
        }
    
        // Write the matrix to device memory each 
        // thread writes one element
        out[i * nm + j] = value;//w[nq*nl+i*nl + l];
        }
        """

    kernel_dict_forward={
            'kernel': kernel_str,
            'name': 'forward_hankel',
            'functions': [{
                'name': 'apply_weights',
                'dtypes' : [complex,complex,complex,np.int64,np.int64],
                'shapes' : [(nq,nm),forward_weights.shape,(nq,nm),None,None],
                'arg_roles' : ['output','const_input','input','const_input','const_input'],
                'const_inputs' : [None,forward_weights,None,np.int64(nq),np.int64(nm)],
                'global_range' : (nq,nm),
                'local_range' : None
            }]
        }
    
    kernel_dict_inverse={
            'kernel': kernel_str,
            'name': 'inverse_hankel',
            'functions': [{
                'name': 'apply_weights',
                'dtypes' : [complex,complex,complex,np.int64,np.int64],
                'shapes' : [(nq,nm),inverse_weights.shape,(nq,nm),None,None],
                'arg_roles' : ['output','const_input','input','const_input','const_input'],
                'const_inputs' : [None,inverse_weights,None,np.int64(nq),np.int64(nm)],
                'global_range' : (nq,nm),
                'local_range' : None
            }]
        }

    forward_gpu_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_forward)
    inverse_gpu_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_inverse)

    zernike_hankel_transform = Multiprocessing.comm_module.add_gpu_process(forward_gpu_process)
    inverse_zernike_hankel_transform = Multiprocessing.comm_module.add_gpu_process(inverse_gpu_process)
    return zernike_hankel_transform,inverse_zernike_hankel_transform



def generate_spherical_zernike_ht_gpu(w,l_orders):
    forward_weights= w['forward']
    inverse_weights= w['inverse']
    n_radial_points = forward_weights.shape[1]

    l_max=l_orders.max()

    nq = n_radial_points
    nl = len(l_orders)
    nlm = l_max*(l_max+2)+1

    kernel_str = """
    __kernel void
    apply_weights(__global double2* out, 
    __global double2* w, 
    __global double2* rho, 
    long nq,long nlm, long nl)
    {
  
    long i = get_global_id(0); 
    long j = get_global_id(1);
    long l = (long) sqrt((double)j);

 
    // value stores the element that is 
    // computed by the thread
    double2 value = 0;
    // wlm is of shape (sum_q,nq,m) where sum_q = nq-1
    for (int q = 0; q < nq-1; ++q)
    {
    double2 wqql = w[q*nq*nl + i*nl + l];
    double2 rqlm = rho[(q+1)*nlm + j];
    value.x += wqql.x * rqlm.x - wqql.y * rqlm.y;
    value.y += wqql.x * rqlm.y + wqql.y * rqlm.x;
    }
    
    // Write the matrix to device memory each 
    // thread writes one element
    out[i * nlm + j] = value;//w[nq*nl+i*nl + l];
    }
    """

    kernel_dict_forward={
            'kernel': kernel_str,
            'name': 'forward_hankel',
            'functions': [{
                'name': 'apply_weights',
                'dtypes' : [complex,complex,complex,np.int64,np.int64,np.int64],
                'shapes' : [(nq,nlm),forward_weights.shape,(nq,nlm),None,None,None],
                'arg_roles' : ['output','const_input','input','const_input','const_input','const_input'],
                'const_inputs' : [None,forward_weights,None,np.int64(nq),np.int64(nlm),np.int64(nl)],
                'global_range' : (nq,nlm),
                'local_range' : None
            }]
        }
    
    kernel_dict_inverse={
            'kernel': kernel_str,
            'name': 'inverse_hankel',
            'functions': [{
                'name': 'apply_weights',
                'dtypes' : [complex,complex,complex,np.int64,np.int64,np.int64],
                'shapes' : [(nq,nlm),inverse_weights.shape,(nq,nlm),None,None,None],
                'arg_roles' : ['output','const_input','input','const_input','const_input','const_input'],
                'const_inputs' : [None,inverse_weights,None,np.int64(nq),np.int64(nlm),np.int64(nl)],
                'global_range' : (nq,nlm),
                'local_range' : None
            }]
        }

    forward_gpu_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_forward)
    inverse_gpu_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_inverse)

    zernike_hankel_transform = Multiprocessing.comm_module.add_gpu_process(forward_gpu_process)
    inverse_zernike_hankel_transform = Multiprocessing.comm_module.add_gpu_process(inverse_gpu_process)
    return zernike_hankel_transform,inverse_zernike_hankel_transform

def generate_polar_zernike_ht_gpu(w,used_orders):
    forward_weights= w['forward']
    inverse_weights= w['inverse']
    n_radial_points = forward_weights.shape[1]

    m_max=used_orders.max()

    nq = n_radial_points
    nm = m_max*2+1

    kernel_str = """
    __kernel void
    apply_weights(__global double2* out, 
    __global double2* w, 
    __global double2* rho, 
    long nq,long nm)
    {
  
    long i = get_global_id(0); 
    long j = get_global_id(1);
 
    // value stores the element that is 
    // computed by the thread
    double2 value = 0;
    // wlm is of shape (sum_q,nq,m) where sum_q = nq-1
    for (int q = 0; q < nq-1; ++q)
    {
    double2 wqqm = w[q*nq*nm + i*nm + j];
    double2 rqm = rho[(q+1)*nm + j];
    value.x += wqqm.x * rqm.x - wqqm.y * rqm.y;
    value.y += wqqm.x * rqm.y + wqqm.y * rqm.x;
    }
    
    // Write the matrix to device memory each 
    // thread writes one element
    out[i * nm + j] = value;//w[nq*nl+i*nl + l];
    }
    """

    kernel_dict_forward={
            'kernel': kernel_str,
            'name': 'forward_hankel',
            'functions': [{
                'name': 'apply_weights',
                'dtypes' : [complex,complex,complex,np.int64,np.int64],
                'shapes' : [(nq,nm),forward_weights.shape,(nq,nm),None,None],
                'arg_roles' : ['output','const_input','input','const_input','const_input'],
                'const_inputs' : [None,forward_weights,None,np.int64(nq),np.int64(nm)],
                'global_range' : (nq,nm),
                'local_range' : None
            }]
        }
    
    kernel_dict_inverse={
            'kernel': kernel_str,
            'name': 'inverse_hankel',
            'functions': [{
                'name': 'apply_weights',
                'dtypes' : [complex,complex,complex,np.int64,np.int64],
                'shapes' : [(nq,nm),inverse_weights.shape,(nq,nm),None,None],
                'arg_roles' : ['output','const_input','input','const_input','const_input'],
                'const_inputs' : [None,inverse_weights,None,np.int64(nq),np.int64(nm)],
                'global_range' : (nq,nm),
                'local_range' : None
            }]
        }

    forward_gpu_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_forward)
    inverse_gpu_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_inverse)

    zernike_hankel_transform = Multiprocessing.comm_module.add_gpu_process(forward_gpu_process)
    inverse_zernike_hankel_transform = Multiprocessing.comm_module.add_gpu_process(inverse_gpu_process)
    return zernike_hankel_transform,inverse_zernike_hankel_transform




############################################################
###    create spherical weights for sin cos expansion    ###
def generate_sin_cos_weights_spherical(max_order, n_radial_points,expansion_limit = -1,pi_in_q = True,dimensions=3,n_cpus=False):
    orders = np.arange(max_order+1)
    assert 2*n_radial_points-1>max_order,'Not enough radial points (N_r={}) to calculate weights upto order {}. max_order<2*n_radial_points-1 '.format(n_radial_points,max_order)    
    if expansion_limit < 0:
        expansion_limit = 2*(2*n_radial_points - 1)
        log.info('Default sin cos expansion limit = {}'.format(expansion_limit))
    expansion_orders = np.arange(expansion_limit+1)
    radial_ids = np.arange(n_radial_points)
    mp_mode = Multiprocessing.MPMode_Queue(ignore_first_dimension=True)
    integrals = Multiprocessing.comm_module.request_mp_evaluation(sin_cos_integral_worker,mode = mp_mode,input_arrays=[orders,expansion_orders,radial_ids],const_inputs=[n_radial_points],call_with_multiple_arguments = True,n_processes = n_cpus,split_mode = 'modulus')
    integrals = np.array(integrals)
    
    weights = np.zeros((max_order +1,n_radial_points,n_radial_points),dtype = float)
    summands = np.zeros((max_order +1,expansion_limit + 1,n_radial_points,n_radial_points),dtype = float)
    pi = np.pi
    N = n_radial_points
    ks = np.arange(N)[None,None,:,None]
    ns = expansion_orders[None,:,None,None]
    #log.info(integrals)
    summands[::2] = np.cos(pi*ks*ns/N)*integrals[::2,:,None,:]

    #summands[::2] = np.cos(pi*ks*ns/N)*integrals[1::2,:,None,:]

    summands[1::2] = np.sin(pi*ks*ns/N)*integrals[1::2,:,None,:]
    summands[...,0,:]*= 0.5 # c_target radial index ks
    summands[:,0,...]*= 0.5 # c_sin_order ns
    weights = np.sum(summands,axis = 1)*2*n_radial_points**2
    return weights

def sin_cos_integral_worker(sp_orders,sc_orders,q_ids,n_radial_points,**kwargs):
    n_parts = len(sp_orders)
    integrals = np.zeros((n_parts),dtype = float)
    odd_mask = (sp_orders%2).astype(bool)
    #log.info('integrals shape = {}'.format(integrals.shape))
    integrands = get_integrands_sin_cos(n_radial_points,sp_orders,odd_mask,sc_orders,q_ids)
    
    integrals[~odd_mask] = spIntegrate.quad_vec(integrands[0],0,1)[0]
    integrals[odd_mask] = spIntegrate.quad_vec(integrands[1],0,1)[0]
    return integrals
        
def get_integrands_sin_cos(n_radial_points,ls,odd_mask,ns,ps):
    Nr=n_radial_points
    pi=np.pi
    cos=np.cos
    sin=np.sin
    jnu = spherical_bessel_jnu
    even_mask= ~odd_mask
    even_ls = ls[even_mask]
    even_ps = ps[even_mask]
    even_ns = ns[even_mask]
    odd_ls = ls[odd_mask]
    odd_ps = ps[odd_mask]
    odd_ns = ns[odd_mask]
    #log.info('even ls = {}'.format(even_ls))
    #log.info('odd ls = {}'.format(odd_ls))
    def even(x):
        return jnu(even_ls,x*pi*even_ps)*cos(pi*even_ns*x)*x**2
    def odd(x):
        return jnu(odd_ls,x*pi*odd_ps)*sin(pi*odd_ns*x)*x**2
        #return jnu(odd_ls,x*pi*odd_ps)*sin(pi*odd_ns*x)*x**2
    return [even,odd]
    


############################################################
###    create spherical weights for sin cos expansion    ###
def generate_sin_cos_weights_spherical_midpoint(max_order, n_radial_points,expansion_limit = -1,pi_in_q = True,dimensions=3,n_cpus=False):
    orders = np.arange(max_order+1)
    assert 2*n_radial_points-1>max_order,'Not enough radial points (N_r={}) to calculate weights upto order {}. max_order<2*n_radial_points-1 '.format(n_radial_points,max_order)    
    if expansion_limit < 0:
        expansion_limit = 2*(2*n_radial_points - 1)
        log.info('Default sin cos expansion limit = {}'.format(expansion_limit))
    expansion_orders = np.arange(expansion_limit+1)
    radial_ids = np.arange(n_radial_points)+1/2
    mp_mode = Multiprocessing.MPMode_Queue(ignore_first_dimension=True)
    integrals = Multiprocessing.comm_module.request_mp_evaluation(sin_cos_integral_worker,mode = mp_mode,input_arrays=[orders,expansion_orders,radial_ids],const_inputs=[n_radial_points],call_with_multiple_arguments = True,n_processes = n_cpus,split_mode = 'modulus')
    integrals = np.array(integrals)
    
    weights = np.zeros((max_order +1,n_radial_points,n_radial_points),dtype = float)
    summands = np.zeros((max_order +1,expansion_limit + 1,n_radial_points,n_radial_points),dtype = float)
    pi = np.pi
    N = n_radial_points
    ks = np.arange(N)[None,None,:,None]+1/2
    ns = expansion_orders[None,:,None,None]
    #log.info(integrals)
    summands[::2] = np.cos(pi*ks*ns/N)*integrals[::2,:,None,:]
    summands[1::2] = np.sin(pi*ks*ns/N)*integrals[1::2,:,None,:]
    #summands[:,0,...]*= 0.5 # c_sin_order ns original
    summands[::2,0,...]*= 0.5 # in paper
    weights = np.sum(summands,axis = 1)*2*n_radial_points**2
    return weights
    



if os.getcwd()[-6:]!="source":
    #print(os.getcwd())
    os.chdir('../../')

import numpy as np
import traceback

from analysisLibrary.filters import BraggFilter as Q_Mean_Sigma_Cart
from control.communicators import mock_comm_module_single_thread


from database.general import load_settings
from analysisLibrary.regrid import AgipdRegridder

from library.mathLibrary import cartesian_to_spherical
from library.mathLibrary import spherical_to_cartesian

from detectors.agipd import AGIPD
from library import physicsLibrary as pLib
from library import units
from database.euxfel import fxs2510_DB
from database.euxfel import fxs2510_EXP_DB

from library.gridLibrary import GridFactory
from library.gridLibrary import get_linspace_log2
from library.physicsLibrary import pixel_grid_to_scattering_grid
from presenters.openGLPresenter import FramePresenter

import log
log=log.setup_custom_logger('root','INFO')

settings_ana = load_settings('fxs2510','latest',target='analysis')
settings_exp = load_settings('fxs2510','proc_latest',target='experiment')
db_exp = fxs2510_EXP_DB(**settings_exp['IO'])
db_ana = fxs2510_DB(**settings_ana['IO'])

agipd = AGIPD(db_exp,load_geometry_file=True)
agipd.origin=np.array([0,0,1500])
d_shape = (16,512,128)
old_data = np.arange(np.prod(d_shape)).reshape(d_shape)
energy = 6e3
wavelength = pLib.energy_to_wavelength(energy)

p_grid = pixel_grid_to_scattering_grid(agipd.pixel_grid,wavelength,out_coord_sys = 'cartesian')
p_grid *= units.standardLength
comm = mock_comm_module_single_thread()

data_grid = p_grid[:,:-1,:-1][agipd.sensitive_pixel_mask].reshape(agipd.data_shape+(3,))

opt = {'data_grid' : data_grid, 'mode':'mask'}

bragg_filter = Q_Mean_Sigma_Cart(opt,comm)


n = 20
datasets = [np.full(agipd.data_shape,1e-3) for i in range(n)]
bragg_centers = [(np.random.rand(3)*np.array([16,512,128])).astype(int) for i in range(int(n/2))]

for i,center in enumerate(bragg_centers):
    data = datasets[i]
    data[center[0],center[1]-1:center[1]+2,center[2]-1:center[2]+2] = 1.2

datasets = np.array(datasets)
pixel_mask = np.zeros(agipd.pixel_grid.shape[:-1],dtype = bool)
pixel_mask[:,:-1,:-1] = agipd.sensitive_pixel_mask
p = FramePresenter(pixel_vertices = p_grid.reshape(-1,3),pixel_indices = agipd.pixel_corner_index.flatten(), sensitive_pixel_mask = pixel_mask)

#datasets = [mask.astype(np.float32) for mask in bragg_filter.radial_masks]

#p.show(datasets)
result = bragg_filter.mask_bragg_pixels(np.array(datasets),bragg_filter.radial_masks,3,1e-3,False)
expected_bragg_count = np.array([9]*int(n/2)+[0]*(n-int(n/2)))
bragg_count = np.sum(result[1].reshape(result[1].shape[0],-1),axis = 1)
if ~(result[0]!=1e-03).any() and (bragg_count == expected_bragg_count).all():
    print('test_passed')
else:
    print('Failed')



import logging
import sys
import os
import numpy as np
import traceback
from scipy.stats import unitary_group
file_path = os.path.realpath(__file__)
plugin_dir = os.path.dirname(file_path)
os.chdir(plugin_dir)


#from analysisLibrary.classes import ReciprocalProjectionData
from .projectLibrary import fxs_invariant_tools as i_tools
from .projectLibrary.ft_grid_pairs import max_order_from_n_angular_steps,get_grid,get_polar_fft_angles_from_max_order
from .projectLibrary.fourier_transforms import generate_ft,load_fourier_transform_weights
from .projectLibrary.harmonic_transforms import HarmonicTransform
from .projectLibrary.hankel_transforms import generate_weightDict
from .projectLibrary.misk import generate_calc_center
from .projectLibrary.fxs_Projections import generate_shift_by_operator
from xframe.library.pythonLibrary import xprint
from xframe.library.gridLibrary import SampledFunction,NestedArray,GridFactory
from xframe.library.mathLibrary import nearest_positive_semidefinite_matrix
from xframe.library.mathLibrary import polar_spherical_dft_reciprocity_relation_radial_cutoffs,distance_from_line_2d
from xframe.interfaces import ProjectWorkerInterface
from xframe import database,settings
from xframe.library.mathLibrary import SampleShapeFunctions
from .projectLibrary.misk import _get_reciprocity_coefficient
from xframe import Multiprocessing
from xframe.library.mathLibrary import spherical_to_cartesian
log=logging.getLogger('root')

#class Worker(RecipeInterface):
class ProjectWorker(ProjectWorkerInterface):
    def __init__(self):
        self.extractor = InvariantExtractor()
    #       global db
    #        db = database.project
    def run(self):
        db = database.project
        opt = settings.project
        
        self.extractor.extract()
            
        if self.extractor.success:
            xprint('Saving results')
            db.save('ccd', self.extractor.cc_data,model_density = self.extractor.density,grid=self.extractor.grid_pair.realGrid[:])
        return {},locals()


######### InvariantExtractor #########
class InvariantExtractor:
    def __init__(self):
        log.info('xyx')
        self.dimensions = False
        self.xray_wavelength = False
        self.enforce_cross_correlation_qq_sym = False
        self.enforce_cross_correlation_phi_sym = False    
        self.average_intensity = False
        self.integrated_intensity = False
        self.data_projection_matrices = {'I1I1':False}
        self.data_low_resolution_intensity_coefficients = False
        self.pi_in_q = None
        self.reciprocity_coefficient = np.pi
        self.max_order = settings.project.grid.max_order
        self.data_n_particles = False
        self.data_radial_points = False
        self.data_angular_points = False
        self.data_max_q = False
        self.data_min_q = False
        self.proj_min_q = {'I1I1':False}
        self.number_of_particles = 1
        self.extraction_routines = {
            'shapes': self.extract_bl_from_shapes,
            'density': self._bl_from_density
        }
        self.b_coeff={'I1I1':False}
        self.b_coeff_masks={'I1I1':False}
        self.b_coeff_q_id_limits = {'I1I1':False}
        self.data_projection_matrices_masks = {'I1I1':False}
        self.data_projection_matrices_q_id_limits = {'I1I1':False}
        self.data_projection_matrix_error_estimates={'I1I1':False}
        self.success=False

            
    def set_standard_instance_variables(self):
        opt = settings.project
        self.dimensions = opt.dimensions
        

    #################################
    ###    extraction routines    ###
    def extract_bl_from_shapes(self):        
        opt = settings.project        
        shape_opt = opt
        
        if shape_opt.GPU.use:
            settings.general.n_control_workers = shape_opt.GPU.n_gpu_workers
            Multiprocessing.comm_module.restart_control_worker()
        self.reciprocity_coefficient = _get_reciprocity_coefficient(shape_opt.fourier_transform)        
        #self.pi_in_q = shape_opt.fourier_transform.pi_in_q
        
        # create_grid
        centers = np.asarray(shape_opt['shapes']['centers'])
        log.info('centers = {}'.format(centers))
        sizes = np.asarray(shape_opt['shapes']['sizes'])
        types = np.asarray(shape_opt['shapes']['types'])
        density_values = np.asarray(shape_opt['shapes']['densities'])
        random_orientation = np.asarray(shape_opt['shapes']['random_orientation'])
        max_particle_radius = shape_opt.get('shape_size','not given')
        if not isinstance(max_particle_radius,(float,int)):
            max_particle_radius = np.max(centers[:,0]+sizes)
        else:
            max_particle_radius/=2 #since mag radius is half the size of the shape.
            
        oversampling = shape_opt.grid.oversampling        
        if isinstance(shape_opt.grid.max_q,bool):
            max_r = oversampling*max_particle_radius
            n_radial_points = shape_opt.grid.n_radial_points
            max_q = polar_spherical_dft_reciprocity_relation_radial_cutoffs(max_r,n_radial_points,reciprocity_coefficient = self.reciprocity_coefficient)
        else:
            max_q = shape_opt.grid.max_q#*(self.reciprocity_coefficient/np.pi)
            n_radial_points = shape_opt.grid.n_radial_points
            max_r = polar_spherical_dft_reciprocity_relation_radial_cutoffs(max_q,n_radial_points,self.reciprocity_coefficient)
        self.data_max_q = max_q
        log.info('maxR = {} maxQ ={} reciprocity_coefficient = {}'.format(max_r,max_q,self.reciprocity_coefficient))

        ht_opt = {'dimensions':self.dimensions,'max_order':self.max_order,**shape_opt.grid}
        log.info('max order = {}'.format(self.max_order))
        cht = HarmonicTransform('complex',ht_opt)
        self.cht = cht
        #weight_dict = generate_weightDict_zernike_spherical(max_order,n_radial_points)        
        grid_opt={'dimensions':opt.dimensions,'type':shape_opt.fourier_transform.type,'max_q':max_q,'n_radial_points':n_radial_points,**cht.grid_param,'reciprocity_coefficient':self.reciprocity_coefficient}
        grid_pair = get_grid(grid_opt)
        self.grid_pair = grid_pair
        if opt.dimensions==3:
            self.data_radial_points =  grid_pair.reciprocalGrid[:,0,0,0]
            #self.data_angular_points = grid_pair.reciprocalGrid[0,0,:,2]
            self.data_angular_points = get_polar_fft_angles_from_max_order(self.max_order) #grid_pair.reciprocalGrid[0,0,:,2]
        elif opt.dimensions==2:
            self.data_radial_points =  grid_pair.reciprocalGrid[:,0,0]
            self.data_angular_points = grid_pair.reciprocalGrid[0,:,1]
        self.data_max_q = np.max(self.data_radial_points)

        log.info('grid shape = {}'.format(grid_pair.realGrid[:].shape))
        log.info(f'grid extends: min {spherical_to_cartesian(grid_pair.realGrid[:]).min()} max {spherical_to_cartesian(grid_pair.realGrid[:]).max()}')

        xprint('Creating density:')
        density = np.zeros(grid_pair.realGrid[:].shape[:-1],dtype = float)
        for shape_type,center,size,dval,rand_rot in zip(types,centers,sizes,density_values,random_orientation):
            log.info(f'\n t {type}\n c {center}\n s {size}\n d {dval}\n r {random_orientation}')
            if shape_type == 'sphere':
                norm = 'standard'
                f = SampleShapeFunctions.get_disk_function(size,lambda points: np.full(points.shape[:-1],dval),center=center,norm=norm,random_orientation=rand_rot,coordSys='spherical')
            elif shape_type == 'tetrahedron':
                if opt.dimensions==3:
                    f = SampleShapeFunctions.get_tetrahedral_function(size,lambda points: np.full(points.shape[:-1],dval),center=center,random_orientation=rand_rot)
                else:
                    f = SampleShapeFunctions.get_polygon_function(size,amplitude_function = lambda points: np.full(points.shape[:-1],dval),center=center,coordSys='polar',random_orientation=rand_rot)
            elif shape_type == 'cube':
                norm = 'inf'
                log.info('size = {}'.format(sizes[i]))
                f = SampleShapeFunctions.get_disk_function(size,lambda points: np.full(points.shape[:-1],dval),center=center,norm=norm,random_orientation=rand_rot)
            else:
                norm = 'inf'
                f = SampleShapeFunctions.get_disk_function(size,lambda points: np.full(points.shape[:-1],dval),center=center,norm=norm,random_orientation=rand_rot)
            density += f(grid_pair.realGrid[:])
        self.density = density
        xprint('done.\n')
        log.info('densty shape = {} grid shape = {} , max_R = {}'.format(density.shape,grid_pair.realGrid[:].shape,grid_pair.realGrid[...,0].max()))
        xprint('Calculating Bl invariants:')
        self._bl_from_density(density,shape_opt,opt.structure_name,n_radial_points,grid_pair,cht)
        xprint('done.\n')

    def generate_fourier_transforms(self, grid_pair, harm_trf):
        opt = settings.project
        dimensions = opt.dimensions
        if dimensions==3:
            r_max = grid_pair.realGrid[:,0,0,0].max()
        elif dimensions ==2:
            r_max = grid_pair.realGrid[:,0,0].max()
        n_radial_points = grid_pair.reciprocalGrid.shape[0]
        
        weights_dict = load_fourier_transform_weights(opt.dimensions,opt.fourier_transform,opt.grid,database.project)

        ft_opt = opt.fourier_transform
        ft_type = ft_opt['type']
        #pi_in_q = ft_opt['pi_in_q']
        reciprocity_coefficient = _get_reciprocity_coefficient(ft_opt)
        use_gpu = opt.GPU.use
        log.info('reciprocity_coefficient = {}'.format(reciprocity_coefficient))
        fourierTransform,inverseFourierTransform=generate_ft(r_max,weights_dict,harm_trf,dimensions,reciprocity_coefficient=reciprocity_coefficient,use_gpu = use_gpu,mode = ft_type)      
        return fourierTransform,inverseFourierTransform

    def _bl_from_density(self,density,opt,name,n_radial_points,grid_pair,cht):
        self.reciprocity_coefficient = _get_reciprocity_coefficient(opt.fourier_transform)        
        db = database.project
        ft,ift = self.generate_fourier_transforms(grid_pair,cht)
        #log.info('start b_coeff calculation')
        #log.info('density shape = {}'.format(density.shape))
        #log.info('grid shape = {}'.format(grid_pair.real.shape))
        
        if settings.project.dimensions ==2:
            grid_type = 'polar'
        elif settings.project.dimensions ==3:
            grid_type = 'spherical'
        
        log.info('max_density = {}'.format(density.max()))
        n_particles=opt.n_particles

        bl = i_tools.density_to_deg2_invariants(density.astype(complex),ft,settings.project.dimensions,cht=cht)
        bl*=self.number_of_particles
        bl[0]*=self.number_of_particles
        self.b_coeff['I1I1'] = bl
        bl_shape = self.b_coeff['I1I1'].shape
        self.b_coeff_masks['I1I1'] = np.full(bl_shape,True)
        q_id_limits = np.zeros((bl_shape[0],)+(2,2),dtype=int)
        q_id_limits[...,1]=bl_shape[0]
        self.b_coeff_q_id_limits={'I1I1': q_id_limits}
        #self.b_coeff['I2I2'] = i_tools.density_to_deg2_invariants((density**2).astype(complex),ft,settings.project.dimensions,cht=cht)
        #self.b_coeff_masks['I2I2'] = np.full(self.b_coeff['I2I2'].shape,True)
        
        #self.b_coeff['I2I1'] = i_tools.density_to_deg2_invariants((density**2).astype(complex),ft,settings.project.dimensions,density2 =density.astype('complex'),cht=cht)
        #self.b_coeff_masks['I2I1'] = np.full(self.b_coeff['I2I1'].shape,True)
        #self.b_coeff[0] *= n_particles
        #log.info(self.b_coeff.shape)

        if settings.project.dimensions ==2:
            self.average_intensity = np.sqrt(np.diag(self.b_coeff['I1I1'][0].real))
        elif settings.project.dimensions ==3:
            self.average_intensity = np.sqrt(np.diag(self.b_coeff['I1I1'][0]).real/(4*np.pi))

    def from_data(self,data):
        self.dimensions = data['dimensions']
        self.xray_wavelength = data['xray_wavelength']
        self.average_intensity = data['average_intensity']
        #log.info('aint type ={}'.format(type(self.average_intensity)))
        self.data_radial_points = data['data_radial_points'][:]
        self.data_angular_points = data['data_angular_points'][:]
        self.data_max_q = np.max(self.data_radial_points)
        self.data_min_q = data.get('data_min_q',False)
        #self.pi_in_q = data.get('pi_in_q',False)
        self.pi_in_q = data.get('pi_in_q',True)
        self.max_order = data['max_order']
        self.number_of_particles=data.get('number_of_particles',1)
        self.data_projection_matrices = data['data_projection_matrices']
        log.info('data keys = {}'.format(data.keys()))
        self.b_coeff = data['b_coeff']
        if isinstance(self.b_coeff,bool):
            self.b_coeff = i_tools.projection_matrices_to_deg2_invariant_3d(self.data_projection_matrices)
        self.success =True

        
    ##################################
    ###   calc cross-correlation   ###

    def calc_cross_correlation(self):
        opt = settings.project
        mode = opt.cross_correlation.method
        data_grid = {'qs':self.data_radial_points,'phis':self.data_angular_points}
        wavelength = opt.cross_correlation.xray_wavelength
        bl = self.b_coeff['I1I1']
        if opt.dimensions ==3:
            cc = i_tools.deg2_invariant_to_cc_3d(bl,wavelength,data_grid,mode= mode,n_processes = opt.n_processes)
        elif opt.dimensions ==2:
            cc = i_tools.deg2_invariant_to_cc_2d(bl,self.cht)
        self.cross_correlation = cc
        

    ##################
    ## main routine ##
    def extract(self):
        opt = settings.project
        db = database.project
        self.set_standard_instance_variables()        
        self.extract_bl_from_shapes()

        xprint('Calculating cross-correlation:')
        self.calc_cross_correlation()
        xprint('done.\n')
        if opt.dimensions==3:
            self.integrated_intensity = np.trapz(self.average_intensity.data * self.data_radial_points**2 , x = self.data_radial_points,axis = 0)*4*np.pi
        elif opt.dimensions==2:
            self.integrated_intensity = np.trapz(self.average_intensity.data * self.data_radial_points , x = self.data_radial_points,axis = 0)*4*np.pi

        cc_data = {}
        cc_data['radial_points']=self.data_radial_points
        cc_data['angular_points']=self.data_angular_points
        cc_data['xray_wavelength']=opt.cross_correlation.xray_wavelength
        cc_data['cross_correlation']= {'I1I1':self.cross_correlation}
        cc_data['average_intensity']=self.average_intensity 
        cc_data['deg_2_invariant']= {"I1I1":self.b_coeff['I1I1']}
        cc_data['number_of_particles']=self.number_of_particles
        
        self.cc_data = cc_data

        self.success =True
        #log.info(self.data_projection_matrices['I1I1'])
        log.info('Extraction completed! \n')
        



import shutil
import os

def copy(source,destination):
    shutil.copyfile(source,destination)

def save(path,data,**kwargs):
    with open(path,'wt') as txt_file:
        if isinstance(data,str):
            txt_file.write(data)
        elif isinstance(data,(list,tuple)):
            txt_file.writelines(data)
  
def load(path,**kwargs):
    with open(path,'rt') as txt_file:
        text = txt_file.readlines()
    return text

def create_path_if_nonexistent(path):
    dirname=os.path.dirname(path)
    if not (os.path.exists(dirname) or dirname==''):
        try:
            os.makedirs(dirname)
        except FileExistsError:
            pass
